#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json

import pytest
from jsonschema import validate
from jsonschema.exceptions import ValidationError

from earthkit.data.utils.paths import earthkit_conf_file

with open(earthkit_conf_file("gridspec_schema.json"), "r") as f:
    schema = json.load(f)


@pytest.mark.parametrize(
    "gridspec",
    [
        {"grid": [0.05, 0.05]},
        {"grid": [2, 2]},
        {"area": [90, -180, -90, 180]},
        {"area": [90, -180, -90, 180], "grid": [2, 2]},
        {"grid": 48},
        {"grid": "F48"},
        {"grid": "N48"},
        {"grid": "O48"},
        {"type": "regular_ll", "grid": [1, 2]},
        {"type": "regular_ll", "area": [90, -180, -90, 180]},
        {"type": "reduced_ll", "grid": 12},
        {"type": "regular_gg", "grid": 48},
        {"type": "regular_gg", "grid": "48"},
        {"type": "regular_gg", "grid": "F48"},
        {"type": "regular_gg", "grid": "F048"},
        {"type": "reduced_gg", "grid": "N48"},
        {"type": "reduced_gg", "grid": "N048"},
        {"type": "reduced_gg", "grid": "O48"},
        {"type": "reduced_gg", "grid": "O048"},
        {
            "type": "mercator",
            "area": [31.173058, 262.036499, 14.736453, 284.975281],
            "grid": [45000.0, 45000.0],
            "lad": 14.0,
            "nx": 56,
            "ny": 44,
            "orientation": 0.0,
        },
        {"grid": "H8", "ordering": "ring"},
        {"type": "healpix", "grid": "H8", "ordering": "ring"},
        {"type": "healpix", "grid": "H8", "ordering": "nested"},
    ],
)
def test_gridspec_schema_valid(gridspec):
    validate(instance=gridspec, schema=schema)


@pytest.mark.parametrize(
    "gridspec",
    [
        {"type": "latlon", "grid": [2, 2]},
        {"type": "reduced_latlon", "grid": 2},
        {"grid": "B48"},
        {"type": "regular_ll", "grid": 48},
        {"type": "regular_ll", "grid": "F48"},
        {"type": "regular_ll", "grid": "a"},
        {"type": "regular_ll", "grid": ["a", "b"]},
        {"type": "regular_ll", "grid": [1]},
        {"type": "regular_ll", "grid": [1, 2, 3]},
        {"type": "regular_gg", "grid": "a"},
        {"type": "regular_gg", "grid": "N48"},
        {"type": "regular_gg", "grid": "O48"},
        {"type": "reduced_gg"},
        {"type": "reduced_gg", "grid": 48},
        {"type": "reduced_gg", "grid": "48"},
        {"type": "reduced_gg", "grid": "F048"},
        {"type": "reduced_gg", "grid": "N"},
        {"grid": "H8", "ordering": "a"},
        {"type": "healpix", "ordering": "ring"},
        {"type": "healpix", "grid": 4, "ordering": "ring"},
    ],
)
def test_gridspec_schema_invalid(gridspec):
    with pytest.raises(ValidationError):
        validate(instance=gridspec, schema=schema)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file


@pytest.mark.cache
@pytest.mark.parametrize(
    "url_suffix,kwargs,num,variables,dim_keys,split_values",
    [
        (
            ["level", "pl.grib"],
            {"time_dim_mode": "raw", "split_dims": ["step"], "dim_name_from_role_name": False},
            2,
            ["2t", "msl", "r", "t"],
            ["date", "time", "levelist"],
            [{"step": 0}, {"step": 6}],
        ),
        (
            ["level", "pl.grib"],
            {
                "time_dim_mode": "raw",
                "split_dims": ["step"],
                "ensure_dims": "step",
                "dim_name_from_role_name": False,
            },
            2,
            ["2t", "msl", "r", "t"],
            ["date", "time", "step", "levelist"],
            [{"step": 0}, {"step": 6}],
        ),
        (
            ["cds-reanalysis-era5-single-levels-20230101-low-resol.grib"],
            {
                "time_dim_mode": "valid_time",
                "split_dims": ["stream", "dataType", "edition", "Ni"],
                "dim_name_from_role_name": False,
            },
            11,
            None,
            ["valid_time"],
            [
                {"stream": "enda", "dataType": "an", "edition": 1, "Ni": 18},
                {"stream": "enda", "dataType": "em", "edition": 1, "Ni": 18},
                {"stream": "enda", "dataType": "es", "edition": 1, "Ni": 18},
                {"stream": "enda", "dataType": "fc", "edition": 1, "Ni": 18},
                {"stream": "enda", "dataType": "fc", "edition": 2, "Ni": 18},
                {"stream": "ewda", "dataType": "an", "edition": 1, "Ni": 12},
                {"stream": "ewda", "dataType": "em", "edition": 1, "Ni": 12},
                {"stream": "ewda", "dataType": "es", "edition": 1, "Ni": 12},
                {"stream": "oper", "dataType": "an", "edition": 1, "Ni": 36},
                {"stream": "oper", "dataType": "fc", "edition": 1, "Ni": 36},
                {"stream": "wave", "dataType": "an", "edition": 1, "Ni": 18},
            ],
        ),
        (
            ["level", "pl.grib"],
            {"time_dim_mode": "raw", "split_dims": ["step"], "dim_name_from_role_name": True},
            2,
            ["2t", "msl", "r", "t"],
            ["date", "time", "level"],
            [{"step": 0}, {"step": 6}],
        ),
        (
            ["level", "pl.grib"],
            {
                "time_dim_mode": "raw",
                "split_dims": ["step"],
                "ensure_dims": "step",
                "dim_name_from_role_name": True,
            },
            2,
            ["2t", "msl", "r", "t"],
            ["date", "time", "step", "level"],
            [{"step": 0}, {"step": 6}],
        ),
        # ({"base_datetime_dim": True}, "param", ["r", "t"], ["levelist"]),
        # ({"squeeze": False}, "param", ["r", "t"], ["time", "step", "levelist"]),
    ],
)
def test_xr_split(url_suffix, kwargs, num, variables, dim_keys, split_values):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data", "xr_engine", *url_suffix))

    dim_keys = dim_keys + ["latitude", "longitude"]
    ds_lst, split_coords_lst = ds_ek.to_xarray(**kwargs)
    assert len(ds_lst) == num
    assert len(split_coords_lst) == len(split_values)

    def dict_to_frozenset_of_kvpairs(d):
        return frozenset(d.items())

    _split_coords_lst = frozenset(map(dict_to_frozenset_of_kvpairs, split_coords_lst))
    _split_values = frozenset(map(dict_to_frozenset_of_kvpairs, split_values))
    assert _split_coords_lst == _split_values
    for ds in ds_lst:
        assert sorted(ds.dims) == sorted(dim_keys)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import xarray
from xarray.backends import BackendEntrypoint

LOG = logging.getLogger(__name__)


class EarthkitBackendEntrypoint(BackendEntrypoint):
    def open_dataset(
        self,
        filename_or_obj,
        source_type="file",
        profile="mars",
        variable_key=None,
        drop_variables=None,
        rename_variables=None,
        mono_variable=None,
        extra_dims=None,
        drop_dims=None,
        ensure_dims=None,
        fixed_dims=None,
        dim_roles=None,
        dim_name_from_role_name=None,
        rename_dims=None,
        dims_as_attrs=None,
        time_dim_mode=None,
        level_dim_mode=None,
        squeeze=None,
        add_valid_time_coord=None,
        decode_times=None,
        decode_timedelta=None,
        add_geo_coords=None,
        attrs_mode=None,
        attrs=None,
        variable_attrs=None,
        global_attrs=None,
        coord_attrs=None,
        add_earthkit_attrs=None,
        rename_attrs=None,
        fill_metadata=None,
        remapping=None,
        flatten_values=None,
        lazy_load=None,
        release_source=None,
        strict=None,
        dtype=None,
        array_module=None,
        array_backend=None,
        errors=None,
    ):
        r"""
        filename_or_obj, str, Path or earthkit object
            Input GRIB file or object to be converted to an xarray dataset.
        profile: str, dict or None
            Provide custom default values for most of the kwargs. Currently, the "mars" and "grid" built-in
            profiles are available, otherwise an explicit dict can
            be used. None is equivalent to an empty dict. When a kwarg is specified it will update
            a default value if it is a dict otherwise it will overwrite it. See: :ref:`xr_profile` for more
            information.
        variable_key: str, None
            Metadata key to specify the dataset variables. It cannot be
            defined as a dimension. Default is "param" (in earthkit-data this is the same as "shortName").
            Only enabled when ``mono_variable`` is False or None.
        drop_variables: str, or iterable of str, None
            A variable or list of variables to drop from the dataset. Default is None. Only used when
            ``variable_key`` is enabled.
        rename_variables: dict, None
            Mapping to rename variables. Default is None. Only used  when
            ``variable_key`` is enabled.
        mono_variable: bool, str, None
            If True or str, the dataset will contain a single variable called "data" (or the value
            of the ``mono_variable`` kwarg when it is a str). If False, the dataset will contain
            one variable for each distinct value of ``variable_key`` metadata key. The default value
            (None) expands to False unless the ``profile`` overwrites it.
        extra_dims:  str, or iterable of str, None
            Define additional dimensions on top of the predefined dimensions. Only enabled when no
            ``fixed_dims`` is specified. Default is None. It can be a single item or a list. Each
            item is either a metadata key, or a dict/tuple defining mapping between the dimension
            name and the metadata key. The whole option can be a dict. E.g.

            .. code-block:: python

                # use key "expver" as a dimension
                extra_dims = "expver"
                # use keys "expver" and "steam" as a dimension
                extra_dims = ["expver", "stream"]
                # define dimensions "expver", mars_stream" and "mars_type" from
                # metadata keys "expver", "stream" and "type"
                extra_dims = [
                    "expver",
                    {"mars_stream": "stream"},
                    ("mars_type", "type"),
                ]
                extra_dims = [
                    {
                        "expver": "expver",
                        "mars_stream": "stream",
                        "mars_type": "type",
                    }
                ]

        drop_dims:  str, or iterable of str, None
            Single or multiple dimensions to be ignored. Default is None.
            Default is None.
        ensure_dims: str, or iterable of str, None
            Dimension or dimensions that should be kept even when ``squeeze=True`` and their size
            is only 1. Default is None.
        fixed_dims: str, or iterable of str, None
            Define all the dimensions to be generated. When used no other dimensions will be created.
            Might be incompatible with other settings. Default is None. It can be a single item or a list.
            Each item is either a metadata key, or a dict/tuple defining mapping between the dimension
            name and the metadata key. The whole option can be a dict. E.g.

            .. code-block:: python

                # use key "step" as a dimension
                fixed_dims = "step"
                # use keys "step" and "levelist" as a dimension
                extra_dims = ["step", "levelist"]
                # define dimensions "step", level" and "level_type" from
                # metadata keys "step", "levelist" and "levtype"
                extra_dims = [
                    "step",
                    {"level": "levelist"},
                    ("level_type", "levtype"),
                ]
                extra_dims = [
                    {"step": "step", "level": "levelist", "level_type": "levtype"}
                ]

        dim_roles: dict, None
            Specify the "roles" used to form the predefined dimensions. The predefined dimensions are
            automatically generated when no ``fixed_dims`` specified and comprise the following
            (in a fixed order):

            - ensemble forecast member dimension
            - temporal dimensions (controlled by ``time_dim_mode``)
            - vertical dimensions (controlled by ``level_dim_mode``)

            ``dim_roles`` is a mapping between the "roles" and the metadata keys representing the roles.
            The possible roles are as follows:

            - "number": metadata key interpreted as ensemble forecast members
            - "date": metadata key interpreted as date part of the "forecast_reference_time"
            - "time": metadata key interpreted as time part of the "forecast_reference_time"
            - "step": metadata key interpreted as forecast step
            - "forecast_reference_time": if not specified or None or empty the forecast reference
              time is built using the "date" and "time" roles
            - "valid_time": if not specified or None or empty the valid time is built using the
              "validityDate" and "validityTime" metadata keys
            - "level": metadata key interpreted as level
            - "level_type": metadata key interpreted as level type

            The default values are as follows:

            .. code-block:: python

                {
                    "number": "number",
                    "date": "dataDate",
                    "time": "dataTime",
                    "step": "step",
                    "forecast_reference_time": None,
                    "valid_date": None,
                    "level": "level",
                    "level_type": "typeOfLevel",
                }

            ``dims_roles`` behaves differently to the other kwargs in the sense that
            it does not override but update the default values. So e.g. to change only "number" in
            the defaults it is enough to specify: "dim_roles={"number": "perturbationNumber"}.
        dim_name_from_role_name: bool, None
            If True, the dimension names are formed from the role names. Otherwise the
            dimension names are formed from the metadata keys specified in ``dim_roles``.
            Its default value (None) expands to True unless the ``profile`` overwrites it.
            Only used when no `fixed_dims`` are specified. *New in version 0.15.0*.
        rename_dims: dict, None
            Mapping to rename dimensions. Default is None.
        dims_as_attrs: str, or iterable of str, None
            Dimension or list of dimensions which should be turned to variable
            attributes if they have only one value for the given variable. Default is None.
        time_dim_mode: str, None
            Define how predefined temporal dimensions are formed. The default is "forecast".
            The possible values are as follows:

            - "forecast": adds two dimensions:

              - "forecast_reference_time": built from the "date" and "time" roles
                (see ``dim_roles``) as np.datetime64 values
              - "step": built from the "step" role. When ``decode_time=True`` the values are
                np.timedelta64
            - "valid_time": adds a dimension called "valid_time" as described by the "valid_time"
              role (see ``dim_roles``). Will contain np.datetime64 values.
            - "raw": the "date", "time" and "step" roles are turned into 3 separate dimensions
        level_dim_mode: str, None
            Define how predefined vertical dimensions are formed. The default is "level".
            The possible values are:

            - "level": adds a single dimension according to the "level" role (see ``dim_roles``)
            - "level_per_type": adds a separate dimensions for each level type based on the
              "level" and "level_type" roles.
            - "level_and_type": Use a single dimension for combined level and type of level.
        squeeze: bool, None
            Remove dimensions which have only one valid value. Not applies to dimensions in
            ``ensure_dims``. Its default value (None) expands
            to True unless the ``profile`` overwrites it.
        add_valid_time_coord: bool, None
            Add the `valid_time` coordinate containing np.datetime64 values to the
            dataset. Only makes effect when ``time_dim_mode`` is not "valid_time". Its default
            value (None) expands to False unless the ``profile`` overwrites it.
        decode_times: bool, None
            If True, decode date and datetime coordinates into datetime64 values. If False, leave
            coordinates representing date-like GRIB keys (e.g. "date", "validityDate") encoded as
            native int values. The default value (None) expands to True unless the ``profile``
            overwrites it.
        decode_timedelta: bool, None
            If True, decode coordinates representing time-like or duration-like GRIB keys
            (e.g. "time", "validityTime", "step") into timedelta64 values. If False, leave time-like
            coordinates encoded as native int values, while duration-like coordinates will be encoded
            as int with the units attached to the coordinate as the "units" attribute.
            If None (default), assume the same value of ``decode_times`` unless the ``profile``
            overwrites it.
        add_geo_coords: bool, None
            If True, add geographic coordinates to the dataset when field values are represented by
            a single "values" dimension. Its default value (None) expands
            to True unless the ``profile`` overwrites it.
        flatten_values: bool, None
            if True, flatten the values per field resulting in a single dimension called
            "values" representing a field. Otherwise the field shape is used to form
            the field dimensions. When the fields are defined on an unstructured grid (e.g.
            reduced Gaussian) or are spectral (e.g. spherical harmonics) this option is
            ignored and the field values are always represented by a single "values"
            dimension.  Its default value (None) expands
            to False unless the ``profile`` overwrites it.
        attrs_mode: str, None
            Define how attributes are generated. Default is "fixed". The possible values are:

            - "fixed": Use the attributes defined in ``variable_attrs`` as variables
              attributes and ``global_attrs`` as global attributes.
            - "unique": Use all the attributes defined in ``attrs``, ``variable_attrs``
              and ``global_attrs``. When an attribute has unique value for a dataset
              it will be a global attribute, otherwise it will be a variable attribute.
              However, this logic is only applied if a unique variable attribute can be
              a global attribute according to the CF conventions Appendix A. (e.g. "units" cannot
              be a global attribute). Additionally, keys in ``variable_attrs`` are always used as
              variable attributes, while keys in ``global_attrs`` are always used as global attributes.
        attrs: str, number, callable, dict or list of these, None
            Attribute or list of attributes. Only used when ``attrs_mode`` is ``unique``.
            Its default value (None) expands to [] unless the ``profile`` overwrites it.
            The following attributes are supported:

            - str: Name of the attribute used as a metadata key to generate the value of
              the attribute. Can also be specified by prefixing with "key=" (e.g. "key=level").
              When prefixed with "namespace=" it specifies a metadata namespace
              (e.g. "namespace=parameter"), which will be added as a dict to the attribute.
            - callable: A callable that takes a Metadata object and returns a dict of attributes
            - dict: A dictionary of attributes with the keys as the attribute names.
              If the value is a callable it takes the attribute name and a Metadata object and
              returns the value of the attribute. A str value prefixed with "key=" or
              "namespace=" is interpreted as explained above. Any other values are used as the
              pre-defined value for the attribute.
        variable_attrs: str, number, callable, dict or list of these, None
            Variable attribute or attributes. For the allowed values see ``attrs``. Its
            default value (None) expands to [] unless the ``profile`` overwrites it.
        global_attrs: str, number, dict or list of these, None
            Global attribute or attributes. For the allowed values see ``attrs``. Its
            default value (None) expands to [] unless the ``profile`` overwrites it.
        coord_attrs: dict, None
            To be documented. Default is None.
        add_earthkit_attrs: bool, None
            If True, add earthkit specific attributes to the dataset. Its default value
            (None) expands to True unless the ``profile`` overwrites it.
        rename_attrs: dict, None
            A dictionary of attribute to rename. Default is None.
        fill_metadata: dict, None
            Define fill values to metadata keys. Default is None.
        remapping: dict, None
            Define new metadata keys for indexing. Default is None.
        lazy_load: bool, None
            If True, the resulting Dataset will load data lazily from the
            underlying data source. If False, a DataSet holding all the data in memory
            and decoupled from the backend source will be created.
            Using ``lazy_load=False`` with ``release_source=True`` can provide optimised
            memory usage in certain cases. The default value of ``lazy_load`` (None)
            expands to True unless the ``profile`` overwrites it.
        release_source: bool, None
            Only used when ``lazy_load=False``. If True, memory held in the input fields are
            released as soon as their values are copied into the resulting Dataset. This is
            done per field to avoid memory spikes. The release operation is currently
            only supported for GRIB fields stored entirely in memory, e.g. when read from a
            :ref:`stream <streams>`. When a field does not support the release operation, this
            option is ignored. Having run :obj:`to_xarray` the input data becomes unusable,
            so use this option carefully. The default value of ``release_source`` (None) expands
            to False unless the ``profile`` overwrites it.
        strict: bool, None
            If True, perform stricter checks on hypercube consistency. Its default value (None) expands
            to False unless the ``profile`` overwrites it.
        dtype: str, numpy.dtype or None
            Typecode or data-type of the array data.
        array_backend: str, array namespace, ArrayBackend, None
            The array backend/namespace to use for array operations. The default value (None) is
            expanded to "numpy".
        """
        fieldlist = self._fieldlist(filename_or_obj, source_type)

        if hasattr(fieldlist, "_ek_builder"):
            builder = fieldlist._ek_builder
            return builder.build()
        else:
            from .builder import SingleDatasetBuilder

            if array_module is not None:
                import warnings

                warnings.warn("'array_module' is deprecated. Use 'array_backend' instead", DeprecationWarning)
                if array_backend is None:
                    array_backend = array_module

            _kwargs = dict(
                variable_key=variable_key,
                drop_variables=drop_variables,
                rename_variables=rename_variables,
                mono_variable=mono_variable,
                extra_dims=extra_dims,
                drop_dims=drop_dims,
                ensure_dims=ensure_dims,
                fixed_dims=fixed_dims,
                rename_dims=rename_dims,
                dim_roles=dim_roles,
                dim_name_from_role_name=dim_name_from_role_name,
                dims_as_attrs=dims_as_attrs,
                time_dim_mode=time_dim_mode,
                level_dim_mode=level_dim_mode,
                squeeze=squeeze,
                attrs_mode=attrs_mode,
                attrs=attrs,
                variable_attrs=variable_attrs,
                global_attrs=global_attrs,
                coord_attrs=coord_attrs,
                add_earthkit_attrs=add_earthkit_attrs,
                rename_attrs=rename_attrs,
                add_valid_time_coord=add_valid_time_coord,
                add_geo_coords=add_geo_coords,
                flatten_values=flatten_values,
                fill_metadata=fill_metadata,
                remapping=remapping,
                decode_times=decode_times,
                decode_timedelta=decode_timedelta,
                lazy_load=lazy_load,
                release_source=release_source,
                strict=strict,
                dtype=dtype,
                array_backend=array_backend,
                errors=errors,
            )

            return SingleDatasetBuilder(fieldlist, profile, from_xr=True, backend_kwargs=_kwargs).build()

    @classmethod
    def guess_can_open(cls, filename_or_obj):
        try:
            from earthkit.data.core import Base

            if isinstance(filename_or_obj, Base):
                return True
            elif isinstance(filename_or_obj, str):
                from earthkit.data.readers.grib import is_grib_file

                return is_grib_file(filename_or_obj)
        except Exception:
            LOG.debug(
                "Failed to guess if %s can be opened by the earthkit backend", filename_or_obj, exc_info=True
            )

        return False

    @staticmethod
    def _fieldlist(filename_or_obj, source_type):
        import os
        import pathlib

        from earthkit.data.core import Base

        if isinstance(filename_or_obj, Base):
            ds = filename_or_obj
        elif isinstance(filename_or_obj, (str, os.PathLike, pathlib.Path)):
            from earthkit.data import from_source

            ds = from_source(source_type, filename_or_obj)
        return ds


class XarrayEarthkit:
    def to_fieldlist(self):
        from earthkit.data.indexing.fieldlist import FieldArray

        return FieldArray([f for f in self._to_fields()])

    def to_target(self, target, *args, **kwargs):
        from earthkit.data.targets import to_target

        to_target(target, *args, data=self._obj, **kwargs)

    def to_grib(self, filename):
        import warnings

        warnings.warn(
            "The `to_grib` is deprecated in 0.15.0 and will be removed in a future version. "
            "Use `to_target` instead.",
            DeprecationWarning,
        )
        from earthkit.data.targets import create_target

        with create_target("file", filename) as target:
            for f in self._to_fields():
                target.write(f)

    def _generator(self):
        from earthkit.data import FieldList

        class GeneratorFieldList(FieldList):
            def __init__(self, data):
                self._data = data

            def mutate(self):
                return self

            def __iter__(self):
                return self._data

            def default_encoder(self):
                return "grib"

        return GeneratorFieldList(self._to_fields())


@xarray.register_dataarray_accessor("earthkit")
class XarrayEarthkitDataArray(XarrayEarthkit):
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    @property
    def metadata(self):
        md = self._obj.attrs.get("_earthkit", dict())
        if "message" in md:
            data = md["message"]
            from earthkit.data.readers.grib.memory import GribMessageMemoryReader
            from earthkit.data.readers.grib.metadata import StandAloneGribMetadata
            from earthkit.data.readers.grib.metadata import WrappedMetadata

            handle = next(GribMessageMemoryReader(data)).handle
            bpv = md.get("bitsPerValue", 0)
            res_md = StandAloneGribMetadata(handle)
            if bpv is not None and bpv > 0:
                return WrappedMetadata(res_md, extra={"bitsPerValue": bpv})
            else:
                return res_md

        raise ValueError(
            (
                "Could not generate earthkit metadata from xarray object."
                "Attribute '_earthkit' is missing or contains incorrect data."
            )
        )

    def _remove_earthkit_attrs(self):
        """Create a copy of the dataarray and remove earthkit attributes."""
        da = self._obj
        if "_earthkit" in da.attrs:
            da = da.copy()
            del da.attrs["_earthkit"]
        return da

    def _to_fields(self):
        from .grib import data_array_to_fields

        for f in data_array_to_fields(self._obj, metadata=self.metadata):
            yield f

    def to_netcdf(self, *args, **kwargs):
        """Remove earthkit attributes before writing to netcdf."""
        ds = self._obj
        if "_earthkit" in self._obj.attrs:
            ds = self._remove_earthkit_attrs()

        return ds.to_netcdf(*args, **kwargs)

    def to_device(self, device, *args, array_backend=None, **kwargs):
        """Return a **new** DataArray whose data live on *device*."""
        from earthkit.utils.array import to_device

        moved = to_device(self._obj.data, device, *args, array_backend=array_backend, **kwargs)
        da = self._obj.copy(deep=False)
        da.data = moved
        return da

    @property
    def grid_spec(self):
        """Return the grid specification of the DataArray."""
        try:
            return self.metadata.gridspec
        except Exception:
            return None


@xarray.register_dataset_accessor("earthkit")
class XarrayEarthkitDataSet(XarrayEarthkit):
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def _to_fields(self):
        from .grib import data_array_to_fields

        for var in self._obj.data_vars:
            for f in data_array_to_fields(self._obj[var]):
                yield f

    def _remove_earthkit_attrs(self):
        """Create a copy of the dataset and remove earthkit attributes."""
        ds = self._obj.copy()
        for var in ds.data_vars:
            if "_earthkit" in ds[var].attrs:
                del ds[var].attrs["_earthkit"]

        return ds

    def to_netcdf(self, *args, **kwargs):
        """Remove earthkit attributes before writing to netcdf."""
        ds = self._obj
        for var in self._obj.data_vars:
            if "_earthkit" in self._obj[var].attrs:
                ds = self._remove_earthkit_attrs()
                break

        return ds.to_netcdf(*args, **kwargs)

    def to_device(self, device, *args, array_backend=None, **kwargs):
        """Return a new Dataset with every data variable on the specified ``device``."""
        from earthkit.utils.array import to_device

        ds = self._obj.copy(deep=False)
        for name, var in ds.data_vars.items():
            ds[name].data = to_device(var.data, device, *args, array_backend=array_backend, **kwargs)
        return ds

    @property
    def grid_spec(self):
        """Return the grid specification of the DataSet."""
        try:
            # return grid spec of the first data variable
            var = list(self._obj.data_vars.values())[0]
            return var.earthkit.grid_spec

        except Exception:
            return None



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import pytest

from earthkit.data.arguments.earthkit_types import BoundingBoxType
from earthkit.data.arguments.earthkit_types import DateListType
from earthkit.data.arguments.earthkit_types import DateType
from earthkit.data.arguments.earthkit_types import EnumListType
from earthkit.data.arguments.earthkit_types import EnumType
from earthkit.data.arguments.earthkit_types import FloatListType
from earthkit.data.arguments.earthkit_types import FloatType
from earthkit.data.arguments.earthkit_types import IntListType
from earthkit.data.arguments.earthkit_types import IntType
from earthkit.data.arguments.earthkit_types import StrListType
from earthkit.data.arguments.earthkit_types import StrType
from earthkit.data.arguments.earthkit_types import VariableListType
from earthkit.data.arguments.earthkit_types import VariableType
from earthkit.data.arguments.transformers import FormatTransformer
from earthkit.data.arguments.transformers import TypeTransformer
from earthkit.data.utils.bbox import BoundingBox

enum = ("a", "b", "c")


def test_types():
    assert TypeTransformer(None, type=EnumType(enum)).transform("a") == "a"

    assert TypeTransformer(None, type=EnumListType(enum)).transform("a") == ["a"]

    assert TypeTransformer(None, type=StrType).transform(42) == "42"
    assert TypeTransformer(None, type=StrListType).transform(42) == ["42"]

    assert TypeTransformer(None, type=IntType).transform("42") == 42
    assert TypeTransformer(None, type=IntListType).transform("42") == [42]

    assert TypeTransformer(None, type=FloatType).transform("3.14") == 3.14
    assert TypeTransformer(None, type=FloatListType).transform(3.14) == [3.14]

    assert TypeTransformer(None, type=DateType).transform(20000101) == datetime.datetime(2000, 1, 1)

    assert TypeTransformer(None, type=DateListType).transform("20000101/to/20000103") == [
        datetime.datetime(2000, 1, 1),
        datetime.datetime(2000, 1, 2),
        datetime.datetime(2000, 1, 3),
    ]

    assert TypeTransformer(None, type=DateListType).transform((20000101, 20000102, 20000103)) == [
        datetime.datetime(2000, 1, 1),
        datetime.datetime(2000, 1, 2),
        datetime.datetime(2000, 1, 3),
    ]

    with pytest.raises(AssertionError):  # FIXME: Not sure what this should be
        assert TypeTransformer(None, type=VariableType("cf")).transform(42) == 0

    with pytest.raises(AssertionError):  # FIXME: Not sure what this should be
        assert TypeTransformer(None, type=VariableListType("cf")).transform(42) == 0

    assert TypeTransformer(None, type=BoundingBoxType).transform((1, -1, -1, 1)) == BoundingBox(
        north=1, west=-1, south=-1, east=1
    )


def test_formats():
    assert FormatTransformer(None, type=EnumType(enum), format="%4s").transform("a") == "   a"

    assert FormatTransformer(None, type=EnumListType(enum), format="%4s").transform(("a", "b")) == [
        "   a",
        "   b",
    ]

    assert FormatTransformer(None, type=StrType, format="%4s").transform("a") == "   a"

    assert FormatTransformer(None, type=StrListType, format="%4s").transform(("a", "b")) == ["   a", "   b"]

    assert FormatTransformer(None, type=IntType, format="%04d").transform(42) == "0042"

    assert FormatTransformer(None, type=IntListType, format="%04d").transform((42, 43)) == ["0042", "0043"]

    assert FormatTransformer(None, type=FloatType, format="%4s").transform("3.14") == "3.14"

    assert FormatTransformer(None, type=FloatType, format="%.1f").transform(3.14) == "3.1"

    assert FormatTransformer(None, type=FloatListType, format="%.1f").transform((3.14, 2.72)) == [
        "3.1",
        "2.7",
    ]

    assert (
        FormatTransformer(None, type=DateType, format="%Y").transform(datetime.datetime(2000, 1, 1)) == "2000"
    )

    assert FormatTransformer(None, type=DateListType, format="%d").transform(
        (datetime.datetime(2000, 1, 1), datetime.datetime(2000, 1, 2))
    ) == ["01", "02"]

    with pytest.raises(Exception):  # FIXME: Not sure what this should be
        assert FormatTransformer(None, type=VariableType, format="%4s").transform(42) == 0

    with pytest.raises(Exception):  # FIXME: Not sure what this should be
        assert FormatTransformer(None, type=VariableListType, format="%4s").transform(42) == 0

    with pytest.raises(Exception):  # FIXME: Not sure what this should be
        assert FormatTransformer(None, type=BoundingBoxType, format="%4s").transform(
            (1, -1, -1, 1)
        ) == BoundingBox(north=1, west=-1, south=-1, east=1)

    b1 = BoundingBox(north=90, west=-45, south=-90, east=45)
    assert FormatTransformer(None, type=BoundingBoxType, format=tuple).transform(b1) == (
        90.0,
        -45.0,
        -90.0,
        45.0,
    )

    assert FormatTransformer(None, type=BoundingBoxType, format="dict").transform(b1) == {
        "east": 45.0,
        "north": 90.0,
        "south": -90.0,
        "west": -45.0,
    }

    assert FormatTransformer(None, type=BoundingBoxType, format=list).transform(b1) == [
        90.0,
        -45.0,
        -90.0,
        45.0,
    ]


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import pathlib
import shutil
from contextlib import contextmanager
from importlib import import_module
from unittest.mock import patch

from earthkit.utils.testing import get_array_backend

from earthkit.data import from_object
from earthkit.data import from_source
from earthkit.data.readers.text import TextReader
from earthkit.data.sources.empty import EmptySource
from earthkit.data.sources.mars import StandaloneMarsClient

LOG = logging.getLogger(__name__)


class OfflineError(Exception):
    pass


_NETWORK_PATCHER = patch("socket.socket", side_effect=OfflineError)

_REMOTE_TEST_DATA_URL = "https://get.ecmwf.int/repository/test-data/earthkit-data/"

_ROOT_DIR = top = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
if not os.path.exists(os.path.join(_ROOT_DIR, "tests", "data")):
    _ROOT_DIR = "./"


@contextmanager
def network_off():
    try:
        _NETWORK_PATCHER.start()
        yield None
    finally:
        _NETWORK_PATCHER.stop()


@contextmanager
def preserve_cwd():
    current_dir = os.getcwd()
    try:
        yield None
    finally:
        os.chdir(current_dir)


def earthkit_remote_test_data_file(*args):
    return os.path.join(_REMOTE_TEST_DATA_URL, *args)


def earthkit_file(*args):
    return os.path.join(_ROOT_DIR, *args)


def earthkit_examples_file(*args):
    return os.path.join(_ROOT_DIR, "docs", "examples", *args)


def earthkit_test_data_file(*args):
    return os.path.join(_ROOT_DIR, "tests", "data", *args)


def data_file(*args):
    return os.path.join(os.path.dirname(__file__), "data", *args)


def file_url(path):
    return pathlib.Path(os.path.abspath(path)).as_uri()


def data_file_url(*args):
    return file_url(data_file(*args))


def modules_installed(*modules):
    for module in modules:
        try:
            import_module(module)
        except ImportError:
            return False
    return True


NO_MARS_DIRECT = not StandaloneMarsClient.enabled()
NO_MARS_API = not os.path.exists(os.path.expanduser("~/.ecmwfapirc"))
NO_MARS = NO_MARS_API and NO_MARS_DIRECT

NO_CDS = not os.path.exists(os.path.expanduser("~/.cdsapirc"))
NO_HDA = not os.path.exists(os.path.expanduser("~/.hdarc"))
IN_GITHUB = os.environ.get("GITHUB_WORKFLOW") is not None
try:
    import ecmwf.opendata  # noqa

    NO_EOD = False
except Exception:
    NO_EOD = True

try:
    import pyfdb  # noqa

    NO_FDB = False
except Exception:
    NO_FDB = True

NO_PROD_FDB = NO_FDB
if not NO_PROD_FDB:
    fdb_home = os.environ.get("FDB_HOME", None)
    NO_PROD_FDB = fdb_home is None


NO_POLYTOPE = not os.path.exists(os.path.expanduser("~/.polytopeapirc"))
NO_COVJSONKIT = not modules_installed("covjsonkit")
NO_RIOXARRAY = not modules_installed("rioxarray")

NO_S3_AUTH = not modules_installed("aws_requests_auth")
NO_GEO = not modules_installed("earthkit-data")
try:
    NO_ECFS = not os.path.exists(shutil.which("ecp"))
except Exception:
    NO_ECFS = True


NO_ZARR = True
try:
    import zarr  # noqa

    if int(zarr.__version__.split(".")[0]) >= 3:
        NO_ZARR = False
except Exception:
    pass


def MISSING(*modules):
    return not modules_installed(*modules)


UNSAFE_SAMPLES_URL = "https://github.com/jwilk/traversal-archives/releases/download/0"


def empty(ds):
    LOG.debug("%s", ds)
    assert isinstance(ds, EmptySource)
    assert len(ds) == 0


def text(ds):
    LOG.debug("%s", ds)
    assert isinstance(ds._reader, TextReader)


UNSAFE_SAMPLES = (
    ("absolute1", empty),
    ("absolute2", empty),
    ("relative0", empty),
    ("relative2", empty),
    ("symlink", text),
    ("dirsymlink", text),
    ("dirsymlink2a", text),
    ("dirsymlink2b", text),
)


def check_unsafe_archives(extension):
    for archive, check in UNSAFE_SAMPLES:
        LOG.debug("%s.%s", archive, extension)
        ds = from_source("url", f"{UNSAFE_SAMPLES_URL}/{archive}{extension}")
        check(ds)


def load_nc_or_xr_source(path, mode):
    if mode == "nc":
        return from_source("file", path)
    else:
        import xarray

        return from_object(xarray.open_dataset(path))


# Array backends
ARRAY_BACKENDS = get_array_backend(["numpy", "torch", "cupy", "jax"], raise_on_missing=False)


def make_tgz(target_dir, target_name, paths):
    import tarfile

    with tarfile.open(os.path.join(target_dir, target_name), "w:gz") as tar:
        for p in paths:
            tar.add(p)
    # tar.close()


def make_zip(target_dir, target_name, paths):
    import zipfile

    with zipfile.ZipFile(os.path.join(target_dir, target_name), "a") as zipf:
        for p in paths:
            zipf.write(p)


WRITE_TO_FILE_METHODS = ["target", "save", "write"]


def write_to_file(mode, path, ds, **kwargs):
    if mode == "target":
        bits_per_value = kwargs.pop("bits_per_value", None)
        md = kwargs.get("metadata", {})
        if bits_per_value is not None:
            md.update({"bitsPerValue": bits_per_value})
            kwargs["metadata"] = md

        ds.to_target("file", path, **kwargs)
    elif mode == "save":
        ds.save(path, **kwargs)
    elif mode == "write":
        append = kwargs.pop("append", False)
        flag = "wb" if not append else "ab"
        with open(path, flag) as f:
            ds.write(f, **kwargs)
    else:
        raise ValueError(f"Invalid {mode=}")


def check_array(
    v,
    shape=None,
    first=None,
    last=None,
    meanv=None,
    eps=1e-3,
    array_backend=None,
):
    if array_backend is None:
        from earthkit.utils.array import get_backend

        array_backend = get_backend(v)

    v = array_backend.to_numpy(v)

    import numpy as np

    assert v.shape == shape
    assert np.isclose(v[0], first, eps)
    assert np.isclose(v[-1], last, eps)
    assert np.isclose(v.mean(), meanv, eps)


def main(path):
    import sys

    import pytest

    # Parallel does not work on darwin, gets RuntimeError: context has already been set
    # because pytest-parallel changes the context from `spawn` to `fork`

    args = ["-p", "no:parallel", "-E", "release"]

    if len(sys.argv) > 1 and sys.argv[1] == "--no-debug":
        args += ["-o", "log_cli=False"]
    else:
        logging.basicConfig(level=logging.DEBUG)
        args += ["-o", "log_cli=True"]

    args += [path]

    sys.exit(pytest.main(args))



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.encoders.grib import GribEncoder
from earthkit.data.targets import to_target
from earthkit.data.testing import earthkit_examples_file


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"encoder": "grib"},
        {"encoder": GribEncoder()},
    ],
)
@pytest.mark.parametrize("direct_call", [True, False])
def test_target_file_pattern_grib_core(kwargs, direct_call):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    with temp_directory() as tmp:
        path = os.path.join(tmp, "{shortName}.grib")

        if direct_call:
            to_target("file-pattern", path, data=ds, **kwargs)
        else:
            ds.to_target("file-pattern", path, **kwargs)

        for name in ("msl", "2t"):
            path = os.path.join(tmp, f"{name}.grib")
            assert os.path.exists(path)
            ds1 = from_source("file", path)
            assert len(ds1) == 1
            assert ds1.metadata("shortName") == [name]


@pytest.mark.parametrize(
    "pattern,expected_value",
    [
        ("{shortName}", {"t": 2, "u": 2, "v": 2}),
        ("{param}", {"t": 2, "u": 2, "v": 2}),
        ("{mars.param}", {"130.128": 2, "131.128": 2, "132.128": 2}),
        ("{shortName}_{level}", {"t_1000": 1, "t_850": 1, "u_1000": 1, "u_850": 1, "v_1000": 1, "v_850": 1}),
        ("{date}_{time}_{step}", {"20180801_1200_0": 6}),
        ("{date}_{time}_{step:03}", {"20180801_1200_000": 6}),
    ],
)
def test_target_file_pattern_grib_keys(pattern, expected_value):
    ds = from_source("file", earthkit_examples_file("test6.grib"))

    with temp_directory() as tmp:
        path = os.path.join(tmp, f"{pattern}.grib")

        to_target(
            "file-pattern",
            path,
            data=ds,
        )

        for k, count in expected_value.items():
            path = os.path.join(tmp, f"{k}.grib")
            assert os.path.exists(path)
            assert len(from_source("file", path)) == count


def test_target_file_pattern_grib_metadata():
    ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))

    with temp_directory() as tmp:
        # setting GRIB keys for the output
        ds.to_target(
            "file-pattern",
            os.path.join(tmp, "{shortName}_{level}.grib"),
            metadata={"date": 20250108},
            bitsPerValue=8,
        )

        for p in ["t", "u", "v"]:
            path = os.path.join(tmp, f"{p}_850.grib")
            assert os.path.exists(path)
            ds1 = from_source("file", path)
            assert len(ds1) == 1
            assert ds1.metadata("shortName") == [p]
            assert ds1.metadata("level") == [850]
            assert ds1.metadata("date") == [20250108]
            assert ds1.metadata("bitsPerValue") == [8]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import time
from collections import defaultdict


class DiagCore:
    def __init__(self, name=""):
        self.name = name

    def _build_label(self, label):
        if label:
            label = f"[{label:10}] "

        if self.name:
            name = f"[{self.name:10}] "
            if label:
                label = f"{name}{label}"
            else:
                label = f"{name}"

        return label


class TimeDiag(DiagCore):
    def __init__(self, name="", **kwargs):
        self.start = time.time()
        self.prev = self.start
        super().__init__(name=name, **kwargs)

    def elapsed(self):
        return time.time() - self.start

    def __call__(self, label="", as_str=False):
        curr = time.time()
        delta = curr - self.prev
        self.prev = curr
        label = self._build_label(label)
        s = f"{label}elapsed={self.elapsed():.3f}s delta={delta:.3f}s"

        if as_str:
            return s
        else:
            print(s)


class MemoryDiag(DiagCore):
    def __init__(self, name="", peak=False, **kwargs):
        import os
        import platform

        import psutil

        self.proc = psutil.Process()
        self.prev = 0
        self.scale = 1
        self.add_peak = peak

        try:
            if os.name == "posix" and platform.system() == "Darwin":
                self.scale = 1.0 / (1024 * 1024)
        except Exception:
            pass

        super().__init__(name=name, **kwargs)

    def scale_to_mbytes(self, v):
        return v * self.scale

    def current(self):
        """Current rss memory usage in MB."""
        import gc

        gc.collect()
        attrs = ["memory_info"]
        mem = self.proc.as_dict(attrs=attrs).get("memory_info", None)
        if mem is not None:
            return self.scale_to_mbytes(mem._asdict()["rss"])
        return 0

    def peak(self):
        """Peak rss memory usage in MB."""
        from resource import RUSAGE_SELF
        from resource import getrusage

        rss = getrusage(RUSAGE_SELF).ru_maxrss
        return self.scale_to_mbytes(rss)

    def __call__(self, label="", delta=True, as_str=False):
        m = self.current()
        _delta = m - self.prev
        self.prev = m
        label = self._build_label(label)

        s = ""
        if delta:
            s = f"{label}curr={m:.3f}MB delta={_delta:.3f}MB"
        else:
            s = f"{label}curr={m:.3f}MB"

        if self.add_peak:
            s += f" peak={self.peak():.3f}MB"

        if as_str:
            return s
        else:
            print(s)


class Diag(DiagCore):
    def __init__(self, name="", peak=False, **kwargs):
        self.time = TimeDiag("")
        self.memory = MemoryDiag("", peak=peak)
        super().__init__(name=name, **kwargs)

    def __call__(self, label=""):
        label = str(label)
        label = self._build_label(label)
        s = f"{label}{self.time(as_str=True)} {self.memory(as_str=True)}"
        print(s)

    def peak(self):
        return self.memory.peak()


def metadata_cache_diag(fieldlist):
    r = defaultdict(int)
    for f in fieldlist:
        collect_field_metadata_cache_diag(f, r)
    return r


def collect_field_metadata_cache_diag(field, r):
    try:
        md_cache = field_cache_diag(field)
        for k in ["metadata_cache_hits", "metadata_cache_misses", "metadata_cache_size"]:
            r[k] += md_cache[k]
    except Exception:
        pass


def field_cache_diag(field):
    r = defaultdict(int)
    try:
        md_cache = field.metadata()._cache
        r["metadata_cache_size"] += len(md_cache)
        r["metadata_cache_hits"] += md_cache.hits
        r["metadata_cache_misses"] += md_cache.misses
    except Exception:
        pass
    return r



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import os
import sys

import pytest

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from forcings_fixtures import load_forcings_fs  # noqa: E402


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
def test_forcings_datetime(input_data):
    ds, _ = load_forcings_fs(last_step=12, input_data=input_data)

    ref = {
        "base_time": [None],
        "valid_time": [
            datetime.datetime(2020, 5, 13, 18),
            datetime.datetime(2020, 5, 14, 0),
        ],
    }

    assert ds.datetime() == ref


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
def test_forcings_valid_datetime(input_data):
    ds, _ = load_forcings_fs(last_step=12, input_data=input_data)
    f = ds[4]

    assert f.metadata("valid_datetime") == "2020-05-13T18:00:00"


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import math

import numpy as np

LOG = logging.getLogger(__name__)


# TODO: refactor this when earthkit.geo grid support is implemented
class Grid:
    def __init__(self, field):
        self.field = field

    @staticmethod
    def make(field):
        # NOTE: underscore grid types are coming from UserMetadata
        grid_type = field.metadata("gridType", default=None)

        if grid_type in ["regular_ll", "_regular_ll"]:
            return RegularLLGrid(field)
        elif grid_type in ["regular_gg", "mercator", "_rectified_ll"]:
            return RectifiedLLGrid(field)
        elif grid_type in ["sh"]:
            return SpectralGrid(field)
        elif grid_type is None or grid_type == "none":
            return NonGrid(field)
        elif grid_type == "_unstructured":
            return Grid(field)
        else:
            return Grid(field)

    def to_distinct_latlon(self, field_shape):
        return None, None

    def to_latlon(self, field_shape=None):
        ll = self.field.to_latlon(flatten=True)
        lat = np.atleast_1d(ll["lat"])
        lon = np.atleast_1d(ll["lon"])

        if field_shape is not None:
            lat = lat.reshape(field_shape)
            lon = lon.reshape(field_shape)
        return lat, lon

    def is_spectral(self):
        return False


class RegularLLGrid(Grid):
    def to_distinct_latlon(self, field_shape):
        assert len(field_shape) == 2
        assert field_shape == self.field.shape
        lat, lon = self.to_latlon()
        _, nx = field_shape
        lat = lat[::nx]
        lon = lon[:nx]
        return lat, lon


class RectifiedLLGrid(Grid):
    def to_distinct_latlon(self, field_shape):
        geo = self.field.metadata().geography
        lat = np.atleast_1d(geo.distinct_latitudes())
        if len(lat) == field_shape[0]:
            lon = np.atleast_1d(geo.distinct_longitudes())
            if len(lon) == field_shape[1]:
                return lat, lon
        return None, None


class SpectralGrid(Grid):
    def to_latlon(self, field_shape=None):
        return None, None

    def is_spectral(self):
        return True


class NonGrid(Grid):
    def to_latlon(self, field_shape=None):
        return None, None

    def is_spectral(self):
        return False


class TensorGrid:
    def __init__(self, field, flatten_values=False):
        self.dims, self.coords, self.coords_dim = self.build(field, flatten_values)

    @staticmethod
    def build(field, flatten_values):
        field_shape = field.shape

        if flatten_values:
            field_shape = (math.prod(field_shape),)

        assert isinstance(field_shape, tuple), (
            field_shape,
            field,
        )

        coords = {}
        dims = {}
        coords_dim = {}

        grid = Grid.make(field)
        if grid.is_spectral():
            if len(field_shape) == 1:
                dims["values"] = field_shape[0]
        else:
            if len(field_shape) == 1:
                dims["values"] = field_shape[0]
                try:
                    lat, lon = grid.to_latlon()
                    if lat is not None and lon is not None:
                        coords["latitude"] = lat
                        coords["longitude"] = lon
                        coords_dim = {k: ("values",) for k in coords}
                except Exception:
                    pass
            elif len(field_shape) == 2:
                try:
                    lat, lon = grid.to_distinct_latlon(field_shape)
                    if (
                        lat is not None
                        and lon is not None
                        and len(lat) == field_shape[0]
                        and len(lon) == field_shape[1]
                    ):
                        coords["latitude"] = lat
                        coords["longitude"] = lon
                        coords_dim["latitude"] = ("latitude",)
                        coords_dim["longitude"] = ("longitude",)
                        dims["latitude"] = lat.size
                        dims["longitude"] = lon.size
                        assert coords["latitude"].size == field_shape[0]
                        assert coords["longitude"].size == field_shape[1]
                        assert dims["latitude"] == field_shape[0]
                        assert dims["longitude"] == field_shape[1]
                except Exception as e:
                    print(e)
                    pass

                if not coords or not dims:
                    lat, lon = grid.to_latlon(field_shape)
                    if lat is not None and lon is not None:
                        coords["latitude"] = lat
                        coords["longitude"] = lon
                        coords_dim = {k: ("y", "x") for k in coords}
                        dims["y"] = field_shape[0]
                        dims["x"] = field_shape[1]
                        assert coords["latitude"].shape == field_shape
                        assert coords["longitude"].shape == field_shape
            else:
                raise ValueError(f"Unsupported field shape {field_shape}")

        # if hasattr(field, "unload"):
        #     field.unload()

        for k, v in coords.items():
            assert k in coords_dim, f"{k=}, {coords_dim=}"
            assert all(x in dims for x in coords_dim[k]), f"{k=}, {coords_dim=} {dims=}"
            assert v.size == math.prod([dims[x] for x in coords_dim[k]])

        return dims, coords, coords_dim



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from . import EncodedData
from . import Encoder

LOG = logging.getLogger(__name__)


class ZarrEncodedData(EncodedData):
    def __init__(self, ds):
        self.ds = ds

    def to_bytes(self):
        return None

    def to_file(self, f):
        return None

    def to_xarray(self):
        return self.ds

    def metadata(self, key):
        raise NotImplementedError


class ZarrEncoder(Encoder):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def encode(
        self,
        data=None,
        **kwargs,
    ):
        if data is not None:
            from earthkit.data.wrappers import get_wrapper

            # data = get_wrapper(data, fieldlist=False)
            data = get_wrapper(data)

            if hasattr(data, "_encode"):
                return data._encode(self, **kwargs)
            else:
                return data
        else:
            raise ValueError("No data to encode")

    def _encode(
        self,
        data=None,
        values=None,
        min=None,
        max=None,
        check_nans=False,
        metadata={},
        template=None,
        # return_bytes=False,
        missing_value=9999,
        **kwargs,
    ):
        return ZarrEncodedData(data.to_xarray(add_earthkit_attrs=False))

    def _encode_field(self, field, **kwargs):
        raise NotImplementedError("ZarrEncoder does not support encoding individual fields.")

    def _encode_fieldlist(self, data, **kwargs):
        earthkit_to_xarray_kwargs = kwargs.pop("earthkit_to_xarray_kwargs", {})
        # earthkit_to_xarray_kwargs.update(kwargs)
        earthkit_to_xarray_kwargs["add_earthkit_attrs"] = False
        kwargs = earthkit_to_xarray_kwargs

        ds = data.to_xarray(**kwargs)
        return ZarrEncodedData(ds)

    def _encode_xarray(self, data, **kwargs):
        raise NotImplementedError


encoder = ZarrEncoder



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from grib_fixtures import FL_TYPES  # noqa: E402
from grib_fixtures import load_grib_data  # noqa: E402


@pytest.mark.parametrize("fl_type", FL_TYPES)
@pytest.mark.parametrize(
    "index,expected_meta",
    [
        (0, ["t", 1000]),
        (2, ["v", 1000]),
        (17, ["v", 300]),
        (-1, ["v", 300]),
        (-5, ["u", 400]),
    ],
)
def test_grib_single_index(fl_type, index, expected_meta):
    f, _ = load_grib_data("tuv_pl.grib", fl_type)
    # f = from_source("file", earthkit_examples_file("tuv_pl.grib"))

    r = f[index]
    assert r.metadata(["shortName", "level"]) == expected_meta
    v = r.values
    assert v.shape == (84,)
    # eps = 0.001
    # assert np.isclose(v[1088], 304.5642, eps)


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_single_index_bad(fl_type):
    f, _ = load_grib_data("tuv_pl.grib", fl_type)
    with pytest.raises(IndexError):
        f[27]


@pytest.mark.parametrize("fl_type", FL_TYPES)
@pytest.mark.parametrize(
    "indexes,expected_meta",
    [
        (slice(0, 4), [["t", 1000], ["u", 1000], ["v", 1000], ["t", 850]]),
        (slice(None, 4), [["t", 1000], ["u", 1000], ["v", 1000], ["t", 850]]),
        (slice(2, 9, 2), [["v", 1000], ["u", 850], ["t", 700], ["v", 700]]),
        (slice(8, 1, -2), [["v", 700], ["t", 700], ["u", 850], ["v", 1000]]),
        (slice(14, 18), [["v", 400], ["t", 300], ["u", 300], ["v", 300]]),
        (slice(14, None), [["v", 400], ["t", 300], ["u", 300], ["v", 300]]),
    ],
)
def test_grib_slice_single_file(fl_type, indexes, expected_meta):
    f, _ = load_grib_data("tuv_pl.grib", fl_type)
    r = f[indexes]
    assert len(r) == 4
    assert r.metadata(["shortName", "level"]) == expected_meta
    v = r.values
    assert v.shape == (4, 84)
    # check the original fieldlist
    assert len(f) == 18
    assert f.metadata("shortName") == ["t", "u", "v"] * 6


@pytest.mark.parametrize(
    "indexes,expected_meta",
    [
        (slice(1, 4), [["msl", 0], ["t", 500], ["z", 500]]),
        (slice(1, 6, 2), [["msl", 0], ["z", 500], ["z", 850]]),
        (slice(5, 0, -2), [["z", 850], ["z", 500], ["msl", 0]]),
        (slice(3, 6), [["z", 500], ["t", 850], ["z", 850]]),
    ],
)
def test_grib_slice_multi_file(indexes, expected_meta):
    f = from_source(
        "file",
        [earthkit_examples_file("test.grib"), earthkit_examples_file("test4.grib")],
    )
    r = f[indexes]
    assert len(r) == 3
    assert r.metadata(["shortName", "level"]) == expected_meta
    # v = r.values
    # assert v.shape == (3, 84)
    # check the original fieldlist
    assert len(f) == 6
    assert f.metadata("shortName") == ["2t", "msl", "t", "z", "t", "z"]


@pytest.mark.parametrize("fl_type", FL_TYPES)
@pytest.mark.parametrize(
    "indexes1,indexes2",
    [
        (np.array([1, 16, 5, 9]), np.array([1, 3])),
        ([1, 16, 5, 9], [1, 3]),
        ((1, 16, 5, 9), (1, 3)),
    ],
)
def test_grib_array_indexing(fl_type, indexes1, indexes2):
    f, _ = load_grib_data("tuv_pl.grib", fl_type)

    r = f[indexes1]
    assert len(r) == 4
    assert r.metadata("shortName") == ["u", "u", "v", "t"]

    r1 = r[indexes2]
    assert len(r1) == 2
    assert r1.metadata("shortName") == ["u", "t"]


@pytest.mark.skip(reason="Index range checking disabled")
@pytest.mark.parametrize("fl_type", FL_TYPES)
@pytest.mark.parametrize(
    "indexes",
    [
        (np.array([1, 19, 5, 9])),
        ([1, 16, 5, 9], [1, 3]),
        ((1, 16, 5, 9), (1, 3)),
    ],
)
def test_grib_array_indexing_bad(fl_type, indexes):
    f = load_grib_data("tuv_pl.grib", fl_type)
    with pytest.raises(IndexError):
        f[indexes]


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_fieldlist_iterator(fl_type):
    g, _ = load_grib_data("tuv_pl.grib", fl_type)
    sn = g.metadata("shortName")
    assert len(sn) == 18
    iter_sn = [f.metadata("shortName") for f in g]
    assert iter_sn == sn
    # repeated iteration
    iter_sn = [f.metadata("shortName") for f in g]
    assert iter_sn == sn


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_fieldlist_iterator_with_zip(fl_type):
    # test something different to the iterator - does not try to
    # 'go off the edge' of the fieldlist, because the length is determined by
    # the list of levels
    g, _ = load_grib_data("tuv_pl.grib", fl_type)
    ref_levs = g.metadata("level")
    assert len(ref_levs) == 18
    levs1 = []
    levs2 = []
    for k, f in zip(g.metadata("level"), g):
        levs1.append(k)
        levs2.append(f.metadata("level"))
    assert levs1 == ref_levs
    assert levs2 == ref_levs


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_fieldlist_iterator_with_zip_multiple(fl_type):
    # same as test_fieldlist_iterator_with_zip() but multiple times
    g, _ = load_grib_data("tuv_pl.grib", fl_type)
    ref_levs = g.metadata("level")
    assert len(ref_levs) == 18
    for i in range(2):
        levs1 = []
        levs2 = []
        for k, f in zip(g.metadata("level"), g):
            levs1.append(k)
            levs2.append(f.metadata("level"))
        assert levs1 == ref_levs, i
        assert levs2 == ref_levs, i


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_fieldlist_reverse_iterator(fl_type):
    g, _ = load_grib_data("tuv_pl.grib", fl_type)
    sn = g.metadata("shortName")
    sn_reversed = list(reversed(sn))
    assert sn_reversed[0] == "v"
    assert sn_reversed[17] == "t"
    gr = reversed(g)
    iter_sn = [f.metadata("shortName") for f in gr]
    assert len(iter_sn) == len(sn_reversed)
    assert iter_sn == sn_reversed
    assert iter_sn == ["v", "u", "t"] * 6



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os

LOG = logging.getLogger(__name__)


_ROOT_DIR = os.path.dirname(os.path.dirname(__file__))


def earthkit_conf_file(*args):
    return os.path.join(_ROOT_DIR, "conf", *args)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_directory
from earthkit.data.testing import NO_CDS
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import preserve_cwd
from earthkit.data.testing import write_to_file

CDS_TIMEOUT = pytest.CDS_TIMEOUT


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.parametrize("prompt", [True, False])
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_grib_prompt(prompt):
    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        prompt=prompt,
        time="12:00",
    )
    assert len(s) == 2
    assert s.metadata("param") == ["2t", "msl"]


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_grib_kwargs():
    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time="12:00",
    )
    assert len(s) == 2
    assert s.metadata("param") == ["2t", "msl"]


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_grib_dict():
    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        dict(
            variable=["2t", "msl"],
            product_type="reanalysis",
            area=[50, -50, 20, 50],
            date="2012-12-12",
            time="12:00",
        ),
    )
    assert len(s) == 2
    assert s.metadata("param") == ["2t", "msl"]


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_grib_invalid_args_kwargs():
    with pytest.raises(TypeError):
        from_source(
            "cds",
            "reanalysis-era5-single-levels",
            dict(
                variable=["2t", "msl"],
                product_type="reanalysis",
                area=[50, -50, 20, 50],
                date="2012-12-12",
            ),
            time="12:00",
        )


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_grib_split_on_var():
    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time="12:00",
        split_on="variable",
    )
    assert len(s) == 2
    assert s.metadata("param") == ["2t", "msl"]
    assert not hasattr(s, "path")
    assert len(s._indexes) == 2


@pytest.mark.parametrize(
    "date,expected_date",
    [
        # (
        #     [datetime.datetime(2012, 12, 12), datetime.datetime(2012, 12, 14)],
        #     [20121212, 20121212, 20121214, 20121214],
        # ),
        (
            ["20121212/20121214"],
            [20121212, 20121212, 20121213, 20121213, 20121214, 20121214],
        ),
        # (
        #     "2012-12-12/to/2012-12-14",
        #     [20121212, 20121212, 20121213, 20121213, 20121214, 20121214],
        # ),
        (
            "2012-12-12/2012-12-14",
            [20121212, 20121212, 20121213, 20121213, 20121214, 20121214],
        ),
        (
            ["2012-12-12/2012-12-14"],
            [20121212, 20121212, 20121213, 20121213, 20121214, 20121214],
        ),
        (
            ["2014-05-12", "2014-05-13", "2014-05-14"],
            [20140512, 20140512, 20140513, 20140513, 20140514, 20140514],
        ),
    ],
)
@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_grib_multi_var_date(date, expected_date):
    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date=date,
        time="12:00",
    )
    assert len(s) == 6
    assert s.metadata("param") == ["2t", "msl"] * 3
    assert s.metadata("date") == expected_date


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_cds_grib_save(write_method):
    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time="12:00",
    )
    with temp_directory() as tmpdir:
        # Check file save to assigned filename
        write_to_file(write_method, os.path.join(tmpdir, "test.grib"), s)
        assert os.path.isfile(os.path.join(tmpdir, "test.grib"))

    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time="12:00",
    )
    with temp_directory() as tmpdir:
        # Check file can be saved in current dir with detected filename:
        with preserve_cwd():
            os.chdir(tmpdir)
            s.save()
            assert os.path.isfile(os.path.basename(s.source_filename))


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
@pytest.mark.parametrize(
    "split_on,expected_file_num,expected_param,expected_time",
    (
        [None, 1, ["2t", "msl", "2t", "msl"], [0, 0, 1200, 1200]],
        [[], 1, ["2t", "msl", "2t", "msl"], [0, 0, 1200, 1200]],
        [{}, 1, ["2t", "msl", "2t", "msl"], [0, 0, 1200, 1200]],
        ["variable", 2, ["2t", "2t", "msl", "msl"], [0, 1200] * 2],
        [("variable",), 2, ["2t", "2t", "msl", "msl"], [0, 1200, 0, 1200]],
        [{"variable": 1}, 2, ["2t", "2t", "msl", "msl"], [0, 1200, 0, 1200]],
        [{"variable": 1, "time": 2}, 2, ["2t", "2t", "msl", "msl"], [0, 1200, 0, 1200]],
        [{"variable": 2, "time": 1}, 2, ["2t", "msl", "2t", "msl"], [0, 0, 1200, 1200]],
        [("variable", "time"), 4, ["2t", "2t", "msl", "msl"], [0, 1200, 0, 1200]],
        [{"variable": 1, "time": 1}, 4, ["2t", "2t", "msl", "msl"], [0, 1200, 0, 1200]],
    ),
)
def test_cds_split_on(split_on, expected_file_num, expected_param, expected_time):
    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time=["00:00", "12:00"],
        split_on=split_on,
    )

    if expected_file_num == 1:
        assert hasattr(s, "path")
        assert not hasattr(s, "_indexes")
    else:
        assert not hasattr(s, "path")
        assert len(s._indexes) == expected_file_num

    assert len(s) == 4
    assert s.metadata("param") == expected_param
    assert s.metadata("time") == expected_time


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
@pytest.mark.parametrize(
    "split_on1,split_on2,expected_file_num,expected_param,expected_time",
    (
        [None, None, 2, ["2t", "2t", "msl", "msl"], [0, 1200, 0, 1200]],
        [None, "time", 3, ["2t", "2t", "msl", "msl"], [0, 1200, 0, 1200]],
        ["time", "time", 4, ["2t", "2t", "msl", "msl"], [0, 1200, 0, 1200]],
    ),
)
def test_cds_multiple_requests(split_on1, split_on2, expected_file_num, expected_param, expected_time):
    base_request = dict(
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time=["00:00", "12:00"],
    )
    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        {**base_request, **{"variable": "2t", "split_on": split_on1}},
        {**base_request, **{"variable": "msl", "split_on": split_on2}},
    )
    assert len(s._indexes) == expected_file_num
    assert len(s) == 4
    assert s.metadata("param") == expected_param
    assert s.metadata("time") == expected_time


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_netcdf():
    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time="12:00",
        format="netcdf",
    )
    assert len(s) == 2
    assert s.metadata("variable") == ["t2m", "msl"]


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_netcdf_save():
    s = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time="12:00",
        format="netcdf",
    )

    with temp_directory() as tmpdir:
        # Check file can be saved in current dir with detected filename:
        with preserve_cwd():
            os.chdir(tmpdir)
            s.save()
            assert os.path.isfile(os.path.basename(s.source_filename))


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(60)
def test_cds_netcdf_selection_limited():
    s = from_source(
        "cds",
        "satellite-albedo",
        {
            "variable": "albb_bh",
            "satellite": "noaa_7",
            "sensor": "avhrr",
            "product_version": "v2",
            "horizontal_resolution": "4km",
            "year": "1983",
            "month": "01",
            "nominal_day": "10",
        },
    )
    assert len(s) == 9
    assert s.metadata("variable") == [
        "AL_BH_BB",
        "AL_BH_BB_ERR",
        "AL_BH_NI",
        "AL_BH_NI_ERR",
        "AL_BH_VI",
        "AL_BH_VI_ERR",
        "AGE",
        "NMOD",
        "QFLAG",
    ]


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_observation_csv_file_to_pandas_xarray():
    collection_id = "insitu-observations-gruan-reference-network"
    request = {
        "format": "csv-lev.zip",
        "year": "2006",
        "month": "05",
        "variable": ["air_temperature", "altitude"],
        "day": ["21", "22"],
    }
    data_cds = from_source("cds", collection_id, **request)
    data_file = from_source("file", data_cds.path)
    assert "report_timestamp" in data_cds.to_pandas().columns

    # Assert consistent behaviour for local and CDS versions
    assert data_cds.to_pandas().equals(data_file.to_pandas())
    assert data_cds.to_xarray().equals(data_file.to_xarray())

    df = data_file.to_pandas()
    assert len(df) == 11318
    assert list(df.columns)[:2] == ["station_name", "report_timestamp"]


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_non_observation_csv_file_to_pandas_xarray():
    collection_id = "sis-energy-derived-projections"
    request = {
        "format": "zip",
        "variable": "wind_power_generation_onshore",
        "spatial_aggregation": "country_level",
        "energy_product_type": "capacity_factor_ratio",
        "temporal_aggregation": "daily",
        "experiment": "rcp_2_6",
        "rcm": "hirham5",
        "gcm": "ec_earth",
    }
    data_cds = from_source("cds", collection_id, **request)
    assert "Date" in data_cds.to_pandas().columns

    # Assert a consistent behviour for local and remote versions
    data_file = from_source("file", data_cds.path)
    assert data_cds.to_pandas().equals(data_file.to_pandas())
    assert data_cds.to_xarray().equals(data_file.to_xarray())


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_grib_to_pandas_xarray():
    collection_id = "reanalysis-era5-single-levels"
    request = dict(
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12/to/2012-12-15",
        time="12:00",
    )
    data_cds = from_source("cds", collection_id, **request)

    # Assert a consistent behviour for local and remote versions
    data_file = from_source("file", data_cds.path)
    assert data_cds.to_pandas().equals(data_file.to_pandas())
    assert data_cds.to_xarray().equals(data_file.to_xarray())

    df = data_file.to_pandas()
    assert len(df) == 388168
    assert list(df.columns)[:3] == ["lat", "lon", "value"]

    ds = data_file.to_xarray()
    assert len(ds) == 2
    assert len(ds.data_vars) == 2


@pytest.mark.long_test
@pytest.mark.download
@pytest.mark.skipif(NO_CDS, reason="No access to CDS")
@pytest.mark.timeout(CDS_TIMEOUT)
def test_cds_netcdf_to_pandas_xarray():
    collection_id = "satellite-methane"
    request = {
        "format": "zip",
        "processing_level": "level_2",
        "variable": "xch4",
        "sensor_and_algorithm": "merged_emma",
        "version": "4.4",
        "year": "2021",
        "month": "01",
        "day": "01",
    }

    data_cds = from_source("cds", collection_id, **request)
    assert "xch4" in data_cds.to_xarray().data_vars

    # Assert a consistent behviour for local and remote versions
    data_file = from_source("file", data_cds.path)
    assert data_cds.to_xarray().equals(data_file.to_xarray())

    ds = data_file.to_xarray()
    assert len(ds) == 18
    assert len(ds.data_vars) == 18

    # TODO: implement to_dataframe
    # assert data_cds.to_pandas().equals(data_file.to_pandas())


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

try:
    import polytope
except ImportError:
    raise ImportError("Polytope access requires 'polytope-client' to be installed")

from . import Source
from .multi_url import MultiUrl
from .prompt import APIKeyPrompt
from .url import Url

LOG = logging.getLogger(__name__)


class PolytopeWebKeyPrompt(APIKeyPrompt):
    register_or_sign_in_url = ("",)
    retrieve_api_key_url = ("",)
    prompts = [
        dict(name="user_email", title="Your email"),
        dict(
            name="user_key",
            example="b295aad8af30332fad2fa8c963ab7900",
            title="API key",
            hidden=True,
            validate="[0-9a-z]{32}",
        ),
    ]

    rcfile = "~/.polytopeapirc"
    config_env = ("POLYTOPE_USER_EMAIL", "POLYTOPE_USER_KEY")


class Polytope(Source):
    """Retrieve data using the Polytope Web API.
    See polytope-client.readthedocs.io for more information.

    Parameters
    ----------
    dataset : str
        The name of the dataset to query.
    request: dict[str, str]
        A collection of key : value pairs specifying the dataset.

    Examples
    --------
    >>> src = earthkit.data.from_source("polytope", "ecmwf-mars", request)
    >>> src.to_pandas()  # if tabular data
    >>> src.to_xarray()  # if datacube
    """

    def __init__(
        self,
        dataset,
        request,
        stream=True,
        **kwargs,
    ) -> None:
        super().__init__()
        assert isinstance(dataset, str)

        for k in ["group_by", "batch_size"]:
            if k in kwargs:
                raise ValueError(f"Invalid argument '{k}' for Polytope. Deprecated since 0.8.0.")

        self._stream_kwargs = dict()
        for k in ["read_all"]:
            if k in kwargs:
                self._stream_kwargs[k] = kwargs.pop(k)

        self.stream = stream

        self.request = dict(dataset=dataset, request=request)

        # all the kwargs are passed to the client!
        self.client = polytope.api.Client(**kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.request['dataset']}, {self.request['request']})"

    def mutate(self) -> Source:
        pointers = self.client.retrieve(
            self.request["dataset"],
            self.request["request"],
            pointer=True,
            asynchronous=False,
        )

        urls = [p["location"] for p in pointers]
        LOG.debug(f"{urls=}")

        if self.stream:
            return Url(
                urls,
                stream=True,
                **self._stream_kwargs,
            )
        else:
            return MultiUrl(urls, sort_urls=False)


source = Polytope



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

try:
    from hda.api import DataOrderRequest  # noqa
except ImportError:
    raise ImportError("WEkEO access requires 'hda' to be installed")

from earthkit.data.core.thread import SoftThreadPool
from earthkit.data.decorators import normalize

from .file import FileSource
from .wekeo import EXTENSIONS
from .wekeo import ApiClient as WekeoClient
from .wekeo import HDAAPIKeyPrompt

LOG = logging.getLogger(__name__)


class ApiClient(WekeoClient):
    name = "wekeocds"

    def __int__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)

    def retrieve(self, name, request, target=None):
        rq = {"dataset_id": name}
        rq.update(
            {
                _name: (
                    _value
                    if isinstance(_value, list) or _name in ("data_format", "download_format")
                    else [_value]
                )
                for _name, _value in request.items()
            }
        )

        if "area" in request:
            rq.update({"bbox": request["area"]})

        request["request"] = rq

        return super().retrieve(name, request, target)


class WekeoCdsRetriever(FileSource):
    sphinxdoc = """
    WekeoCdsRetriever
    """

    def __init__(self, dataset, *args, prompt=True, **kwargs):
        super().__init__()

        self.prompt = prompt

        assert isinstance(dataset, str)
        if len(args):
            assert len(args) == 1
            assert isinstance(args[0], dict)
            assert not kwargs
            kwargs = args[0]

        requests = self.requests(**kwargs)

        self.client(self.prompt)  # Trigger password prompt before threading

        nthreads = min(self.config("number-of-download-threads"), len(requests))

        if nthreads < 2:
            self.path = [self._retrieve(dataset, r) for r in requests]
        else:
            from earthkit.data.utils.progbar import tqdm

            with SoftThreadPool(nthreads=nthreads) as pool:
                futures = [pool.submit(self._retrieve, dataset, r) for r in requests]

                iterator = (f.result() for f in futures)
                self.path = list(tqdm(iterator, leave=True, total=len(requests)))

    def _retrieve(self, dataset, request):
        def retrieve(target, args):
            self.client(self.prompt).retrieve(args[0], args[1], target)

        return self.cache_file(
            retrieve,
            (dataset, request),
            extension=EXTENSIONS.get(request.get("format"), ".cache"),
        )

    @normalize("date", "date-list(%Y-%m-%d)")
    @normalize("area", "bounding-box(list)")
    def requests(self, **kwargs):
        split_on = kwargs.pop("split_on", None)
        if split_on is None or not isinstance(kwargs.get(split_on), (list, tuple)):
            return [kwargs]

        result = []

        for v in kwargs[split_on]:
            r = dict(**kwargs)
            r[split_on] = v
            result.append(r)

        return result

    @staticmethod
    def client(use_prompt):
        if use_prompt:
            prompt = HDAAPIKeyPrompt()
            prompt.check()

            try:
                return ApiClient()
            except Exception as e:
                # if no rc file is available hda throws
                # ConfigurationError: Missing or incomplete configuration
                if ".hdarc" in str(e) or not prompt.has_config_env():
                    LOG.warning(e)
                    LOG.exception(f"Could not load hda client. {e}")
                    prompt.ask_user_and_save()
                    return ApiClient()
                raise
        else:
            return ApiClient()


source = WekeoCdsRetriever



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data.utils import projections

NO_CARTOPY = False
try:
    import cartopy.crs as ccrs
except ImportError:
    NO_CARTOPY = True


def test_from_proj_string_laea():
    proj_string = "+proj=laea +lat_0=52 +lon_0=10 +x_0=4321000 +y_0=3210000 +ellps=GRS80 +units=m +no_defs"
    projection = projections.Projection.from_proj_string(proj_string)
    assert isinstance(projection, projections.LambertAzimuthalEqualArea)
    assert projection.parameters == {
        "central_latitude": 52.0,
        "central_longitude": 10.0,
        "false_northing": 3210000.0,
        "false_easting": 4321000.0,
    }
    assert projection.globe == {
        "ellipse": "GRS80",
    }


def test_from_proj_string_lcc():
    proj_string = "+proj=lcc +lon_0=-90 +lat_1=33 +lat_2=45"
    projection = projections.Projection.from_proj_string(proj_string)
    assert isinstance(projection, projections.LambertConformal)
    assert projection.parameters == {
        "central_longitude": -90.0,
        "standard_parallels": (33.0, 45.0),
    }


def test_from_cf_grid_mapping_aea():
    grid_mapping = {
        "grid_mapping_name": "albers_conical_equal_area",
        "standard_parallel": 40,
        "longitude_of_central_meridian": 10,
    }
    projection = projections.Projection.from_cf_grid_mapping(**grid_mapping)
    assert isinstance(projection, projections.AlbersEqualArea)
    assert projection.parameters == {
        "central_longitude": 10,
        "standard_parallels": [40],
    }
    assert projection.globe == dict()


@pytest.mark.skipif(NO_CARTOPY, reason="cartopy is not installed")
def test_to_cartopy_crs_laea():
    proj_string = "+proj=laea +lat_0=52 +lon_0=10 +x_0=4321000 +y_0=3210000 +ellps=GRS80 +units=m +no_defs"
    projection = projections.Projection.from_proj_string(proj_string)
    assert projection.to_cartopy_crs() == ccrs.LambertAzimuthalEqualArea(
        central_longitude=10.0,
        central_latitude=52.0,
        false_easting=4321000.0,
        false_northing=3210000.0,
        globe=ccrs.Globe(ellipse="GRS80"),
    )



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from grib_fixtures import FL_TYPES  # noqa: E402
from grib_fixtures import load_grib_data  # noqa: E402


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_indices_base(fl_type):
    ds, _ = load_grib_data("tuv_pl.grib", fl_type)

    ref_full = {
        "class": ["od"],
        "stream": ["oper"],
        "levtype": ["pl"],
        "type": ["an"],
        "expver": ["0001"],
        "date": [20180801],
        "time": [1200],
        "domain": ["g"],
        "number": [0],
        "levelist": [300, 400, 500, 700, 850, 1000],
        "param": ["t", "u", "v"],
    }

    r = ds.indices()
    assert r == ref_full

    ref = {
        "levelist": [300, 400, 500, 700, 850, 1000],
        "param": ["t", "u", "v"],
    }
    r = ds.indices(squeeze=True)
    assert r == ref

    ref = ["t", "u", "v"]
    r = ds.index("param")
    assert r == ref

    ref = [300, 400, 500, 700, 850, 1000]
    ref_full["level"] = ref

    r = ds.index("level")
    assert r == ref

    r = ds.indices()
    assert r == ref_full


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_indices_sel(fl_type):
    ds, _ = load_grib_data("tuv_pl.grib", fl_type)

    ref = {
        "class": ["od"],
        "stream": ["oper"],
        "levtype": ["pl"],
        "type": ["an"],
        "expver": ["0001"],
        "date": [20180801],
        "time": [1200],
        "domain": ["g"],
        "number": [0],
        "levelist": [300, 400, 500, 700, 850, 1000],
        "param": ["t"],
    }

    ds1 = ds.sel(param="t")
    r = ds1.indices()
    assert r == ref

    ref = {
        "levelist": [300, 400, 500, 700, 850, 1000],
    }
    r = ds1.indices(squeeze=True)
    assert r == ref


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_indices_multi(fl_type):
    f1, _ = load_grib_data("tuv_pl.grib", fl_type)
    f2, _ = load_grib_data("ml_data.grib", fl_type, folder="data")
    ds = f1 + f2

    ref = {
        "class": ["od"],
        "stream": ["oper"],
        "levtype": ["ml", "pl"],
        "type": ["an", "fc"],
        "expver": ["0001"],
        "date": [20180111, 20180801],
        "time": [1200],
        "domain": ["g"],
        "number": [0],
        "levelist": [
            1,
            5,
            9,
            13,
            17,
            21,
            25,
            29,
            33,
            37,
            41,
            45,
            49,
            53,
            57,
            61,
            65,
            69,
            73,
            77,
            81,
            85,
            89,
            93,
            97,
            101,
            105,
            109,
            113,
            117,
            121,
            125,
            129,
            133,
            137,
            300,
            400,
            500,
            700,
            850,
            1000,
        ],
        "param": ["lnsp", "t", "u", "v"],
    }

    r = ds.indices()
    assert r == ref


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_indices_multi_sel(fl_type):
    f1, _ = load_grib_data("tuv_pl.grib", fl_type)
    f2, _ = load_grib_data("ml_data.grib", fl_type, folder="data")
    ds = f1 + f2

    ref = {
        "class": ["od"],
        "stream": ["oper"],
        "levtype": ["ml", "pl"],
        "type": ["an", "fc"],
        "expver": ["0001"],
        "date": [20180111, 20180801],
        "time": [1200],
        "domain": ["g"],
        "number": [0],
        "levelist": [93, 500],
        "param": ["t"],
    }

    ds1 = ds.sel(param="t", level=[93, 500])
    r = ds1.indices()
    assert r == ref


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_indices_order_by(fl_type):
    ds, _ = load_grib_data("tuv_pl.grib", fl_type)

    ref = {
        "class": ["od"],
        "stream": ["oper"],
        "levtype": ["pl"],
        "type": ["an"],
        "expver": ["0001"],
        "date": [20180801],
        "time": [1200],
        "domain": ["g"],
        "number": [0],
        "levelist": [300, 400, 500, 700, 850, 1000],
        "param": ["t", "u", "v"],
    }

    ds1 = ds.order_by(levelist="descending")
    r = ds1.indices()
    assert r == ref



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from earthkit.data.readers.netcdf import NetCDFFieldListUrlReader
from earthkit.data.readers.netcdf import NetCDFReader
from earthkit.data.sources import Source


class OpenDAP(Source):
    def __init__(self, url):
        self.url = url

    def mutate(self):
        fs = NetCDFFieldListUrlReader(source, self.url)
        if fs.has_fields():
            return fs
        else:
            return NetCDFReader(source, self.path)


source = OpenDAP



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from xr_engine_fixtures import compare_coords  # noqa: E402
from xr_engine_fixtures import compare_dims  # noqa: E402


@pytest.mark.cache
@pytest.mark.parametrize(
    "_kwargs,dims,coords",
    [
        ({"fixed_dims": ["valid_time", "param"]}, {"valid_time": 732, "param": 1}, {"param": ["2t"]}),
        ({"fixed_dims": ["param", "valid_time"]}, {"param": 1, "valid_time": 732}, {"param": ["2t"]}),
    ],
)
def test_xr_engine_mono_variable_1(_kwargs, dims, coords):
    ds_in = from_source(
        "url", earthkit_remote_test_data_file("test-data", "xr_engine", "date", "t2_1_year.grib")
    )

    ds = ds_in.to_xarray(
        mono_variable=True,
        chunks={"valid_time": 1},
        add_earthkit_attrs=False,
        **_kwargs,
    )

    assert ds is not None

    dims["latitude"] = 3
    dims["longitude"] = 3

    compare_dims(ds, dims, order_ref_var="data", sizes=True)
    compare_coords(ds, coords)


@pytest.mark.cache
@pytest.mark.parametrize(
    "_kwargs,dims,coords",
    [
        ({"fixed_dims": ["valid_time", "param"]}, {"valid_time": 732, "param": 2}, {"param": ["2d", "2t"]}),
        ({"fixed_dims": ["param", "valid_time"]}, {"param": 2, "valid_time": 732}, {"param": ["2d", "2t"]}),
    ],
)
def test_xr_engine_mono_variable_2(_kwargs, dims, coords):
    ds_in = from_source(
        "url", earthkit_remote_test_data_file("test-data", "xr_engine", "date", "t2_td2_1_year.grib")
    )

    ds = ds_in.to_xarray(
        mono_variable=True,
        chunks={"valid_time": 1},
        add_earthkit_attrs=False,
        **_kwargs,
    )

    assert ds is not None

    dims["latitude"] = 3
    dims["longitude"] = 3

    compare_dims(ds, dims, order_ref_var="data", sizes=True)
    compare_coords(ds, coords)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import math

from .diff import DictDiff
from .diff import ListDiff

LOG = logging.getLogger(__name__)


def list_to_str(vals, n=10):
    try:
        size = f"({len(vals)}) "
        if len(vals) <= n:
            return size + str(vals)
        else:
            lst = size + "[" + ", ".join(str(vals[: n - 1])) + "..., " + str(vals[-1]) + "]"
            return lst
    except Exception:
        return vals


def check_coords(var_name, coord_name, coords, ref_coords):
    if coord_name in ref_coords:
        v1 = ref_coords[coord_name].vals
        v2 = coords[coord_name]
        diff = ListDiff.diff(v1, v2, name=coord_name)
        if not diff.same:
            raise ValueError(
                (
                    f'Variable "{var_name}" has inconsistent dimension "{coord_name}" compared '
                    f"to other variables. Expected values: {list_to_str(v1)}, "
                    f"got: {list_to_str(v2)}. {diff.diff_text}"
                )
            )


class CubeChecker:
    def __init__(self, tensor):
        self.tensor = tensor

    def first_diff(self, coord_keys):
        for i, f in enumerate(self.tensor.source):
            t_coords = self.tensor._index_to_coords_value(i, self.tensor)
            f_coords = f.metadata(coord_keys)
            print(f"Checking field[{i}] with coords {t_coords} vs {f_coords}")
            diff = ListDiff.diff(t_coords, f_coords)
            if not diff.same:
                name = ""
                if diff.diff_index != -1:
                    name = coord_keys[diff.diff_index]

                return i, f, t_coords, f_coords, name, diff

    def neighbour_field(self, field_num, index):
        f_other = None
        index_other = None
        if index > 0:
            index_other = index - 1
        elif index < field_num - 1:
            index_other = index + 1
        if index_other is not None:
            f_other = self.tensor.source[index_other]

        return f_other, index_other

    def namespace_diff(self, f, f_other, namespace):
        meta = f.metadata(namespace=namespace)
        meta_other = f_other.metadata(namespace=namespace)
        return DictDiff.diff(meta, meta_other)

    def meta_diff(self, f, f_other, coords_keys):
        f_coords = f.metadata(coords_keys)
        other_coords = f_other.metadata(coords_keys)
        meta = {coords_keys[i]: v for i, v in enumerate(f_coords)}
        meta_other = {coords_keys[i]: v for i, v in enumerate(other_coords)}
        return DictDiff.diff(meta, meta_other)

    def meta(self, f, coords_keys):
        f_coords = f.metadata(coords_keys)
        return {coords_keys[i]: v for i, v in enumerate(f_coords)}

    def check(self, details=False):
        field_num = len(self.tensor.source)
        cube_num = math.prod(self.tensor._user_shape)

        if field_num == cube_num:
            return

        coord_keys = list(self.tensor._user_coords.keys())
        dims = "\n".join([f"{k} {list_to_str(v)}" for k, v in self.tensor._user_coords.items()])
        text_num = (
            f"Input does not form a full hypercube."
            f" Expected number of fields based on the dimensions does not match "
            f"actual number of fields: {cube_num} != {field_num}.\n"
            f"Dimensions: \n {dims}"
        )

        print(text_num)
        print(coord_keys)

        if not details:
            raise ValueError(text_num)
        else:
            # find first differing field
            r = self.first_diff(coord_keys)
            if r is not None:
                index, f, t_coords, f_coords, name, diff = r
                assert index is not None
                assert index >= 0

                text_dims = (
                    f"\nFirst difference from the expected dimensions occurs at field[{index}]"
                    f' for dimension "{name}". Expected {t_coords[diff.diff_index]}'
                    f" got {f_coords[diff.diff_index]}."
                )

                f_other, index_other = self.neighbour_field(field_num, index)
                if f_other is not None:
                    assert index_other is not None
                    assert index_other >= 0
                    assert index_other != index

                    # namespace diff
                    namespaces = ["mars", "ls", "time", "vertical", "geography"]
                    md_diff = {}
                    diff = self.meta_diff(f, f_other, coord_keys)
                    if not diff.same:
                        md_diff["dims"] = diff.diff_text

                    for ns in namespaces:
                        diff = self.namespace_diff(f, f_other, ns)
                        if not diff.same:
                            md_diff[f"namespace={ns}"] = diff.diff_text

                    text_ns = (
                        f"\nComparing field[{index}] with field[{index_other}], the following "
                        "metadata difference was found:\n"
                    )
                    for k, v in md_diff.items():
                        text_ns += f"{k}:\n {v}\n"

                md = {}
                md["dims"] = self.meta(f, coord_keys)
                for ns in namespaces:
                    md["namespace=" + ns] = f.metadata(namespace=ns)

                text_meta = f"\nField[{index}] metadata:\n"
                for k, v in md.items():
                    text_meta += f"{k}:\n {v}\n"

                raise ValueError(f"{text_num}{text_dims}{text_ns}{text_meta}")



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import WRITE_TO_FILE_METHODS
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.testing import write_to_file


def test_grib_from_memory():
    with open(earthkit_test_data_file("test_single.grib"), "rb") as f:
        data = f.read()
        fs = from_source("memory", data)
        assert len(fs) == 1
        sn = fs.metadata("param")
        assert sn == ["2t"]
        assert fs[0].metadata("shortName") == "2t"


@pytest.mark.parametrize("write_method", WRITE_TO_FILE_METHODS)
def test_grib_save_when_loaded_from_memory(write_method):
    with open(earthkit_test_data_file("test_single.grib"), "rb") as f:
        data = f.read()
        fs = from_source("memory", data)
        with temp_file() as tmp:
            write_to_file(write_method, tmp, fs)
            fs_saved = from_source("file", tmp)
            assert len(fs) == len(fs_saved)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
from collections import defaultdict
from collections import namedtuple

SimplePart = namedtuple("SimplePart", ["offset", "length"])


class Part:
    """Represent a file part."""

    def __init__(self, path, offset, length):
        assert path is not None
        self.path = path
        self.offset = offset
        self.length = length

    def __eq__(self, other):
        return self.path == other.path and self.offset == other.offset and self.length == other.length

    @classmethod
    def resolve(cls, parts, directory=None):
        paths = defaultdict(list)
        for i, part in enumerate(parts):
            paths[part.path].append(part)

        for path, bits in paths.items():
            if path.startswith("http://") or path.startswith("https://") or path.startswith("ftp://"):
                # newpath = download_and_cache(
                #     path, parts=[(p.offset, p.length) for p in bits]
                # )
                # newoffset = 0
                # for p in bits:
                #     p.path = newpath
                #     p.offset = newoffset
                #     newoffset += p.length
                raise ValueError("Part: url based paths are not supported")

            elif directory and not os.path.isabs(path):
                for p in bits:
                    p.path = os.path.join(directory, path)

        return parts

    def __repr__(self):
        return f"Part[{self.path},{self.offset},{self.length}]"


def check_urls_and_parts(urls, parts):
    """Check if urls and parts are compatible

    When any of the ``urls`` contain a part ``parts`` must be None.
    """
    if not isinstance(urls, (list, tuple)):
        urls = [urls]

    # a single url as [url, parts] is not allowed
    if (
        len(urls) == 2
        and isinstance(urls[0], str)
        and (urls[1] is None or isinstance(urls[1], (list, tuple)))
    ):
        if parts is not None:
            raise ValueError("Cannot specify parts both as arg and kwarg")
        urls = [urls]

    if any(not isinstance(u, str) for u in urls):
        if parts is not None:
            raise ValueError("Cannot specify parts both as arg and kwarg")

    return urls


def _ensure_parts(parts):
    if parts is None:
        return None
    if parts == [None]:
        return None
    if len(parts) == 2 and isinstance(parts[0], int) and isinstance(parts[1], int):
        parts = [parts]
    parts = [SimplePart(offset, length) for offset, length in parts]
    if len(parts) == 0:
        return None
    return parts


def ensure_urls_and_parts(urls, parts, compress=True):
    if not isinstance(urls, (list, tuple)):
        urls = [urls]

    parts = _ensure_parts(parts)
    result = []
    for v in urls:
        if isinstance(v, (list, tuple)):
            u, p = v
            result.append((u, _ensure_parts(p)))
        else:
            result.append((v, parts))

    urls_and_parts = []
    # Break into ascending order if needed
    for url, parts in result:
        if parts is None:
            urls_and_parts.append((url, None))
            continue

        last = 0
        newparts = []
        for p in parts:
            if p.offset < last:
                if newparts:
                    urls_and_parts.append((url, compress_parts(newparts)))
                    newparts = []
            newparts.append(p)
            last = p.offset
        urls_and_parts.append((url, compress_parts(newparts)))

    return urls_and_parts


def compress_parts(parts):
    last = -1
    result = []
    # Compress and check
    for offset, length in parts:
        assert offset >= 0 and length > 0
        if offset < last:
            raise Exception(
                f"Offsets and lengths must be in order, and not overlapping:"
                f" offset={offset}, end of previous part={last}"
            )
        if offset == last:
            # Compress
            offset, prev_length = result.pop()
            length += prev_length

        result.append((offset, length))
        last = offset + length
    return tuple(SimplePart(offset, length) for offset, length in result)


class PathAndParts:
    compress = None
    sequence = None

    def __init__(self, path, parts):
        """Should not be instantiated directly.
        The public API are the factory methods."""
        self.path = path
        self.parts = parts

    @classmethod
    def from_paths(cls, path, parts):
        path, parts = cls._parse(path, parts, cls.sequence, cls.compress)
        return cls(path, parts)

    def is_empty(self):
        return not (self.parts is not None and any(x is not None for x in self.parts))

    def update(self, path):
        if self.path != path:
            self.path, self.parts = self._parse(path, self.parts, self.sequence, self.compress)

    @staticmethod
    def _parse(paths, parts, sequence, compress):
        """Preprocess paths and parts.

        Parameters
        ----------
        paths: str or list/tuple
            The path(s). When it is a sequence either each
            item is a path (str), or a pair of a path and :ref:`parts <parts>`.
        parts: part,list/tuple of parts or None.
            The :ref:`parts <parts>`.

        Returns
        -------
        str or list of str
            The path or paths.
        SimplePart, list or tuple, None
            The parts (one for each path). A part can be a single
            SimplePart, a list/tuple of SimpleParts or None.

        """
        if parts is None:
            if isinstance(paths, str):
                if sequence:
                    return [paths], [None]
                else:
                    return paths, None
            elif isinstance(paths, (list, tuple)) and all(isinstance(p, str) for p in paths):
                return paths, [None] * len(paths)

        paths = check_urls_and_parts(paths, parts)
        paths_and_parts = ensure_urls_and_parts(paths, parts, compress=compress)

        paths, parts = zip(*paths_and_parts)
        assert len(paths) == len(parts)
        if not sequence and len(paths) == 1:
            return paths[0], parts[0]
        else:
            return paths, parts

    def zipped(self):
        return [(pt, pr) for pt, pr in self]

    def __iter__(self):
        path = self.path
        parts = self.parts
        if isinstance(self.path, str):
            path = [self.path]
            parts = [self.parts]
        return zip(path, parts)

    def sorted(self):
        if not isinstance(self.path, str):
            path, parts = zip(*sorted(zip(self.path, self.parts), key=lambda x: x[0]))
            return self.__class__(path, parts)
        return self.__class__(self.path, self.parts)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


try:
    # NOTE: the `version.py` file must not be present in the git repository
    #   as it is generated by setuptools at install time
    from ._version import __version__
except ImportError:  # pragma: no cover
    # Local copy or not installed with setuptools
    __version__ = "999"

from earthkit.utils import array as array_api

from earthkit.data.translators import transform
from earthkit.data.wrappers import get_wrapper as from_object

from .arguments.transformers import ALL
from .core.caching import CACHE as cache
from .core.config import CONFIG as config
from .core.fieldlist import Field
from .core.fieldlist import FieldList
from .encoders import create_encoder
from .indexing.fieldlist import SimpleFieldList
from .readers.grib.output import new_grib_output
from .sources import Source
from .sources import from_source
from .sources import from_source_lazily
from .sources.array_list import ArrayField
from .targets import create_target
from .targets import to_target
from .utils.examples import download_example_file
from .utils.examples import remote_example_file

settings = config

__all__ = [
    "ALL",
    "array_api",
    "ArrayField",
    "cache",
    "create_encoder",
    "create_target",
    "download_example_file",
    "Field",
    "FieldList",
    "from_source",
    "from_source_lazily",
    "from_object",
    "transform",
    "new_grib_output",
    "remote_example_file",
    "config",
    "settings",
    "SimpleFieldList",
    "Source",
    "to_target",
    "__version__",
]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import threading
from abc import ABCMeta
from abc import abstractmethod
from functools import cached_property

from earthkit.data.utils import ensure_dict
from earthkit.data.utils import ensure_iterable

LOG = logging.getLogger(__name__)


GEO_KEYS = ["md5GridSection"]
MANDATORY_KEYS = GEO_KEYS
IGNORE_ATTRS = ["md5GridSection"]


class RemappingBuilder:
    def __init__(self, remappings, patches=None):
        self.remappings = dict(**ensure_dict(remappings))
        self.patches = dict(**ensure_dict(patches))

    def build(self):
        from earthkit.data.core.order import build_remapping

        return build_remapping(self.remappings, patches=self.patches)

    def add(self, remapping, patches=None):
        self.remappings.update(remapping)
        if patches is not None:
            self.patches.update(patches)


class ProfileConf:
    def __init__(self):
        self._conf = {}
        self._lock = threading.Lock()

    def get(self, name):
        if name not in self._conf:
            self._load(name)
        return self._conf[name]

    def _load(self, name):
        with self._lock:
            if name not in self._conf and name != "defaults":
                here = os.path.dirname(__file__)
                path = os.path.join(here, f"{name}.yaml")
                if os.path.exists(path):
                    import yaml

                    try:
                        with open(path, "r") as f:
                            self._conf[name] = yaml.safe_load(f)
                    except Exception as e:
                        LOG.exception(f"Failed read profile conf {name} from {path}. {e}")
                        raise
                else:
                    raise ValueError(f"Profile {name} not found! path={path}")

    @cached_property
    def defaults(self):
        here = os.path.dirname(__file__)
        path = os.path.join(here, "defaults.yaml")
        if os.path.exists(path):
            import yaml

            try:
                with open(path, "r") as f:
                    return yaml.safe_load(f)
            except Exception as e:
                LOG.exception(f"Failed read profile defaults file {path}. {e}")
                raise
        else:
            raise ValueError(f"Profile defaults not found! path={path}")


PROFILE_CONF = ProfileConf()


class ProfileVariable(metaclass=ABCMeta):
    key = None

    @property
    @abstractmethod
    def is_mono(self):
        pass

    @abstractmethod
    def update(self, ds):
        """
        Parameters
        ----------
        ds: fieldlist
            FieldList object with cached metadata
        """
        pass

    @abstractmethod
    def check(self, profile):
        """
        Check the variable against the profile.
        This is called after the variable has been updated.
        """
        pass


class KeyVariable(ProfileVariable):
    def __init__(self, key, drop, rename_map):
        self.key = key
        self.drop = drop
        self.rename_map = rename_map
        self.variables = []

    @property
    def is_mono(self):
        return False

    def rename(self, v):
        """
        Rename the variable using the rename map.
        If the variable is not in the rename map, return the original name.
        """
        return self.rename_map.get(v, v)

    def update(self, ds):
        """
        Parameters
        ----------
        ds: fieldlist
            FieldList object with cached metadata
        """

        self.variables = ds.index(self.key)

        if self.drop:
            self.variables = [v for v in self.variables if v not in self.drop]

        if not self.variables:
            raise ValueError(f"No metadata values found for variable key {self.key}")

    def check(self, profile):
        assert self.key not in profile.dim_keys
        assert self.key is not None
        assert self.variables
        assert self.key not in profile.dim_keys


class MonoVariable(ProfileVariable):
    def __init__(self, name=None):
        self.name = name or "data"

    @property
    def is_mono(self):
        return True

    def rename(self, v):
        """
        Rename the variable using the rename map.
        If the variable is not in the rename map, return the original name.
        """
        return v

    def update(self, ds):
        """
        Parameters
        ----------
        ds: fieldlist
            FieldList object with cached metadata
        """
        # MonoVariable does not have a key, so no update needed
        pass

    def check(self, profile):
        pass


class Profile:
    USER_ONLY_OPTIONS = ["remapping", "patches", "fill_metadata"]
    DEFAULT_PROFILE_NAME = "mars"

    def __init__(
        self,
        name=None,
        **kwargs,
    ):
        from .attrs import Attrs
        from .dim import DimHandler

        self._kwargs = dict(**kwargs)
        self.name = name
        self.index_keys = []

        patches = dict()

        # defaults
        fill_md = kwargs.pop("fill_metadata", None)
        if fill_md:
            if not isinstance(fill_md, dict):
                raise ValueError("fill_metadata must be a dict!")
            for k, v in fill_md.items():
                if isinstance(v, (str, int, float, bool)):
                    patches[k] = lambda x: x if x is not None else v
                elif isinstance(v, dict) or callable(v):
                    patches[k] = v

        self.remapping = RemappingBuilder(kwargs.pop("remapping", None), patches)

        # variables
        mono_variable = kwargs.pop("mono_variable")
        variable_key = kwargs.pop("variable_key")
        drop_variables = kwargs.pop("drop_variables")
        rename_variables_map = kwargs.pop("rename_variables")
        if mono_variable:
            mono_variable_name = None
            if isinstance(mono_variable, str):
                mono_variable_name = mono_variable
            self.variable = MonoVariable(name=mono_variable_name)
        else:
            if variable_key:
                self.variable = KeyVariable(variable_key, drop_variables, rename_variables_map)
                self.add_keys([self.variable.key])
            else:
                raise ValueError("variable_key must be a non-empty str!")

        # dims
        self.dims = DimHandler(
            self,
            kwargs.pop("extra_dims"),
            kwargs.pop("drop_dims"),
            kwargs.pop("ensure_dims"),
            kwargs.pop("fixed_dims"),
            kwargs.pop("split_dims"),
            kwargs.pop("rename_dims"),
            kwargs.pop("dim_roles"),
            kwargs.pop("dim_name_from_role_name"),
            kwargs.pop("dims_as_attrs"),
            kwargs.pop("time_dim_mode"),
            kwargs.pop("level_dim_mode"),
            kwargs.pop("squeeze"),
        )

        # coordinates
        self.add_valid_time_coord = kwargs.pop("add_valid_time_coord")
        self.add_geo_coords = kwargs.pop("add_geo_coords")

        # attributes
        self.attrs = Attrs(
            self,
            kwargs.pop("attrs_mode"),
            kwargs.pop("attrs"),
            kwargs.pop("variable_attrs"),
            kwargs.pop("global_attrs"),
            kwargs.pop("coord_attrs"),
            kwargs.pop("rename_attrs"),
        )

        self.add_earthkit_attrs = kwargs.pop("add_earthkit_attrs")

        # generic
        self.decode_times = kwargs.pop("decode_times")
        self.decode_timedelta = kwargs.pop("decode_timedelta")
        self.lazy_load = kwargs.pop("lazy_load")
        self.release_source = kwargs.pop("release_source")
        self.direct_backend = kwargs.pop("direct_backend")
        self.strict = kwargs.pop("strict")
        self.errors = kwargs.pop("errors")

        # values
        self.flatten_values = kwargs.pop("flatten_values")
        self.dtype = kwargs.pop("dtype")
        self.array_backend = kwargs.pop("array_backend")

        if "array_module" in kwargs:
            raise ValueError(
                "'array_module' is deprecated. Use 'array_backend' instead. "
                "If you are using 'array_module', please update your code to use 'array_backend'."
            )

        # if self.array_backend == "numpy":
        #     import numpy as np

        #     self.array_module = np

        if kwargs:
            raise ValueError(f"Unsupported options: {kwargs}")

        self.add_keys(ensure_iterable(MANDATORY_KEYS))

        self.prepend_keys(self.dims.split_dims)

        # print("INIT variable key", self.variable_key)
        # print("INIT index_keys", self.index_keys)
        # print("INIT dim_keys", self.dim_keys)

    @staticmethod
    def make(name_or_def, *args, force=False, **kwargs):
        # print("name_or_def", name_or_def)

        if isinstance(name_or_def, Profile):
            if force:
                return name_or_def.copy()
            else:
                return name_or_def

        if name_or_def is None:
            name_or_def = {}

        if isinstance(name_or_def, str):
            conf = PROFILE_CONF.get(name_or_def)
            name = name_or_def
        elif isinstance(name_or_def, dict):
            conf = name_or_def
            name = ""
        else:
            raise ValueError(f"Unsupported type for name_or_def: {type(name_or_def)}")

        return Profile.from_conf(name, conf, *args, **kwargs)

    @classmethod
    def from_conf(cls, name, conf, *args, **kwargs):
        import copy

        kwargs = copy.deepcopy(kwargs)
        opt = copy.deepcopy(PROFILE_CONF.defaults)

        def _deprec_array_module(data):
            """Deprecated: use 'array_backend' instead"""
            if "array_module" in data:
                import warnings

                warnings.warn("'array_module' is deprecated. Use 'array_backend' instead", DeprecationWarning)

                array_module = kwargs.pop("array_module")
                if data.get("array_backend", None) is None:
                    data["array_backend"] = array_module

        _deprec_array_module(kwargs)

        for d in [conf, kwargs]:
            for k, v in d.items():
                if k in PROFILE_CONF.defaults and v is not None:
                    if isinstance(PROFILE_CONF.defaults[k], dict):
                        if not isinstance(v, dict):
                            raise ValueError(f"Expected dict for key {k} in profile {name}")
                        if "__overwrite__" in v:
                            v.pop("__overwrite__")
                            opt[k] = v
                        else:
                            opt[k].update(v)
                    elif isinstance(PROFILE_CONF.defaults[k], list):
                        opt[k] = ensure_iterable(v)
                    else:
                        opt[k] = v
                elif k not in PROFILE_CONF.defaults:
                    if k in cls.USER_ONLY_OPTIONS:
                        opt[k] = v
                    else:
                        raise ValueError(f"Unknown key {k} in profile {name}")

        def _check_type(k, v, t):
            if isinstance(PROFILE_CONF.defaults[k], t):
                if not isinstance(v, t):
                    raise ValueError(f"Expected {t.__name__} for key {k} in profile {name}")
                return True

        for k, v in opt.items():
            if k in PROFILE_CONF.defaults:
                any(_check_type(k, v, t) for t in [dict, list, bool])

        if opt["decode_timedelta"] is None:
            opt["decode_timedelta"] = opt["decode_times"]

        return cls(*args, name=name, **opt)

    @classmethod
    def to_docs(cls, name):
        """Used to generate documentation"""
        import copy

        if name is None:
            conf = {}
        else:
            conf = PROFILE_CONF.get(name)

        opt = copy.deepcopy(PROFILE_CONF.defaults)

        for k, v in conf.items():
            if k in PROFILE_CONF.defaults and v is not None:
                if isinstance(PROFILE_CONF.defaults[k], dict):
                    if not isinstance(v, dict):
                        raise ValueError(f"Expected dict for key {k} in profile {name}")
                    if "__overwrite__" in v:
                        v.pop("__overwrite__")
                        opt[k] = v
                    else:
                        opt[k].update(v)
                elif isinstance(PROFILE_CONF.defaults[k], list):
                    opt[k] = ensure_iterable(v)
                else:
                    opt[k] = v
            elif k not in PROFILE_CONF.defaults:
                if k in cls.USER_ONLY_OPTIONS:
                    opt[k] = v
                else:
                    raise ValueError(f"Unknown key {k} in profile {name}")

        return opt

    def copy(self):
        return Profile(name=self.name, **self._kwargs)

    @property
    def dim_keys(self):
        return self.dims.active_dim_keys

    def add_keys(self, keys):
        self.index_keys += [key for key in keys if key not in self.index_keys]

    def prepend_keys(self, keys):
        if keys:
            self.index_keys = keys + [k for k in self.index_keys if k not in keys]

    # def update_variables(self, ds):
    #     if not self.mono:
    #         self.variables = ds.index(self.variable_key)

    #         if self.drop_variables:
    #             self.variables = [v for v in self.variables if v not in self.drop_variables]

    #         if not self.variables:
    #             raise ValueError(f"No metadata values found for variable key {self.variable_key}")

    def update(self, ds):
        """
        Parameters
        ----------
        ds: fieldlist
            FieldList object with cached metadata
        attributes: dict
            Index keys which has a single (valid) value
        """
        # print("index_keys", self.index_keys)
        # print("dim_keys", self.dim_keys)
        self.variable.update(ds)
        self.dims.update(ds)
        self.variable.check(self)

        # if not self.mono:
        #     self.variables = ds.index(self.variable_key)

        #     if self.drop_variables:
        #         self.variables = [v for v in self.variables if v not in self.drop_variables]

        #     if not self.variables:
        #         raise ValueError(f"No metadata values found for variable key {self.variable_key}")

        #     # variable_keys = VARIABLE_KEYS + [self.variable_key]
        #     # self.dims.remove(variable_keys, others=True)
        #     self.dims.update(ds)
        #     assert self.variable_key not in self.dim_keys

        #     assert self.variable_key is not None
        #     assert self.variables
        #     assert self.variable_key not in self.dim_keys
        # else:
        #     self.dims.update(ds)

        # print("UPDATE variable_key", self.variable_key)
        # print("UPDATE variables", self.variables)
        # print(" -> dim_keys", self.dim_keys)

    @property
    def sort_keys(self):
        if self.variable.key:
            keys = self.dims.split_dims + [self.variable.key]
        else:
            keys = self.dims.split_dims
        keys += [k for k in self.dim_keys if k not in keys]
        return keys

    def rename_dims_map(self):
        return self.dims.rename_dims_map

    # def rename_variable(self, v):
    #     return self.variable.rename_map.get(v, v)

    def rename_dataset_dims(self, dataset):
        return self.dims.rename_dataset_dims(dataset)
